"""Tests for data utils classes and methods."""
import datetime
import hashlib
import json
from typing import Dict, Generator, Union
from unittest.mock import MagicMock, Mock

import numpy as np
import pandas as pd
from pandas._typing import Dtype
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy

from bitfount.data.exceptions import (
    DatabaseMissingTableError,
    DatabaseSchemaNotFoundError,
    DatabaseUnsupportedQueryError,
)
from bitfount.data.helper import convert_epochs_to_steps
from bitfount.data.utils import (
    DatabaseConnection,
    _convert_python_dtypes_to_pandas_dtypes,
    _generate_dtypes_hash,
    _hash_str,
)
from bitfount.types import _Dtypes
from tests.utils.helper import unit_test


@unit_test
def test_convert_epochs_to_steps() -> None:
    """Test converting of epochs to steps is correct."""
    dataloader = MagicMock()
    dataloader.__len__.return_value = 100
    steps = convert_epochs_to_steps(5, dataloader)
    assert steps == 100 * 5


@unit_test
class TestDatabaseConnection:
    """Tests DatabaseConnection class."""

    @fixture
    def mock_engine(self) -> Mock:
        """Returns mock sqlalchemy engine."""
        return Mock(spec=sqlalchemy.engine.base.Engine)

    @fixture(autouse=True)
    def mock_inspector(self, mocker: MockerFixture) -> Generator[Mock, None, None]:
        """Automatically mocks sqlalchemy inspector and yields mocked object."""
        mock_inspector = Mock(
            default_schema_name="public", spec=sqlalchemy.engine.Inspector
        )
        mock_inspector.get_schema_names.return_value = ["public"]
        mock_inspector.get_table_names.return_value = ["test_data"]
        mocker.patch("bitfount.data.utils.inspect", return_value=mock_inspector)
        yield mock_inspector

    def test_create_engine_from_connection_string(self, mocker: MockerFixture) -> None:
        """Tests that a sqlalchemy object can be created from a database URI."""
        mock_create_engine = mocker.patch(
            "bitfount.data.utils.create_engine", autospec=True
        )
        DatabaseConnection(
            "postgresql://localhost:5432/test", query="SELECT * FROM test_data"
        )
        mock_create_engine.assert_called_once_with("postgresql://localhost:5432/test")

    def test_schema_not_found_in_database_raises_value_error(
        self, mock_engine: Mock, mock_inspector: Mock
    ) -> None:
        """Tests that DatabaseSchemaNotFoundError raised if schema not in database."""
        with pytest.raises(DatabaseSchemaNotFoundError):
            DatabaseConnection(mock_engine, db_schema="nonexistent_schema")

        mock_inspector.get_schema_names.assert_called_once()

    def test_query_and_table_names_raises_value_error(self, mock_engine: Mock) -> None:
        """Tests that query and table names can't both be specified."""
        with pytest.raises(ValueError):
            DatabaseConnection(
                mock_engine, query="SELECT * FROM test_data", table_names=["test_data"]
            )

    def test_specified_table_not_found_in_schema_raises_value_error(
        self,
        mock_engine: Mock,
    ) -> None:
        """Tests that DatabaseMissingTableError raised if table not in schema."""
        with pytest.raises(DatabaseMissingTableError):
            DatabaseConnection(mock_engine, table_names=["nonexistent_table"])

    def test_no_tables_found_in_schema_raises_value_error(
        self, mock_engine: Mock, mock_inspector: Mock
    ) -> None:
        """Tests that DatabaseMissingTableError raised if no tables in schema."""
        mock_inspector.get_table_names.return_value = []
        with pytest.raises(DatabaseMissingTableError):
            DatabaseConnection(mock_engine)

    def test_get_default_schema(self, mock_engine: Mock, mock_inspector: Mock) -> None:
        """Tests that default schema is used if none specified."""
        DatabaseConnection(mock_engine)
        mock_inspector.get_table_names.assert_called_once_with(schema="public")

    def test_single_table_name(self, mock_engine: Mock) -> None:
        """Tests that a single table name can be specified."""
        db_conn = DatabaseConnection(mock_engine, table_names=["test_data"])
        assert not db_conn.multi_table

    def test_multiple_table_names(
        self, mock_engine: Mock, mock_inspector: Mock
    ) -> None:
        """Tests that multiple table names can be specified."""
        mock_inspector.get_table_names.return_value = ["test_data", "test_data_2"]
        db_conn = DatabaseConnection(
            mock_engine, table_names=["test_data", "test_data_2"]
        )
        assert db_conn.multi_table

    def test_all_tables_in_schema(
        self, mock_engine: Mock, mock_inspector: Mock
    ) -> None:
        """Tests that connection will default to all tables if none provided."""
        mock_inspector.get_table_names.return_value = [
            "test_data",
            "test_data_2",
            "test_data_3",
        ]
        db_conn = DatabaseConnection(mock_engine)
        assert db_conn.multi_table
        assert db_conn.table_names == ["test_data", "test_data_2", "test_data_3"]

    def test_query(self, mock_engine: Mock) -> None:
        """Tests that query can be specified."""
        db_conn = DatabaseConnection(mock_engine, query="SELECT * FROM test_data")
        assert not db_conn.multi_table
        assert db_conn.query

    def test_into_query_error(self, mock_engine: Mock) -> None:
        """Tests that a query containing into raises error."""
        with pytest.raises(DatabaseUnsupportedQueryError):
            DatabaseConnection(mock_engine, query="SELECT * INTO df")


@unit_test
class TestDataFrameHashing:
    """Tests for generate_dataframe_hash()."""

    @fixture
    def dtypes(self) -> _Dtypes:
        """A test dataframe with data."""
        return {
            "np_test": np.dtype(int),
            "pd_test": pd.core.arrays.integer.Int64Dtype(),
        }

    @fixture
    def dtypes_hash(self) -> str:
        """The expected hash for the dataframe fixture."""
        # The hash is on the DataFrame.dtypes (which returns a Series), so we
        # manually construct the expected matching one.
        dtypes = {
            "np_test": str(np.dtype(int)),
            "pd_test": str(pd.core.arrays.integer.Int64Dtype()),
        }
        str_rep = json.dumps(dtypes, sort_keys=True)
        return hashlib.sha256(str_rep.encode("utf8")).hexdigest()

    @fixture
    def empty_dtypes(self) -> _Dtypes:
        """A test dtype mapping with no data."""
        return {}

    @fixture
    def empty_dtypes_hash(self) -> str:
        """The expected hash of an empty dataframe."""
        # The hash is on the DataFrame.dtypes (which returns a Series), so we
        # manually construct the expected matching one.
        empty_series: Dict = {}
        str_rep: str = str(empty_series)
        return hashlib.sha256(str_rep.encode("utf8")).hexdigest()

    def test_generate_dtypes_hash(self, dtypes: _Dtypes, dtypes_hash: str) -> None:
        """Tests generated hash is expected one for non-empty dataframe."""
        assert _generate_dtypes_hash(dtypes) == dtypes_hash

    def test_generate_dtypes_hash_empty_dataframe(
        self, empty_dtypes: _Dtypes, empty_dtypes_hash: str
    ) -> None:
        """Tests generated hash is expected one for empty dataframe."""
        assert _generate_dtypes_hash(empty_dtypes) == empty_dtypes_hash

    def test_generate_dtypes_hash_same_for_same_dtypes(
        self, dtypes: Dict[str, Union[Dtype, np.dtype]], dtypes_hash: str
    ) -> None:
        """Tests generated hash is consistent for two dataframes with same cols."""
        dtypes_2 = dtypes.copy()

        # Check they are different instances
        assert dtypes is not dtypes_2
        # Check hashes match
        assert (
            _generate_dtypes_hash(dtypes)
            == _generate_dtypes_hash(dtypes_2)
            == dtypes_hash
        )

    def test_generate_dtypes_hash_different_for_different_dtype_dataframes(
        self, dtypes: _Dtypes, dtypes_hash: str
    ) -> None:
        """Tests hash is different for two dataframes with diff col dtypes."""
        # Change the column dtype from int64 to string
        dtypes_2 = {k: np.dtype(str) for k in dtypes.keys()}

        # Check they are different instances
        assert dtypes is not dtypes_2
        # Check hashes differ
        assert (
            _generate_dtypes_hash(dtypes)
            != _generate_dtypes_hash(dtypes_2)
            != dtypes_hash
        )


@unit_test
def test_hash_str() -> None:
    """Tests that hash_str() works."""
    test_string = "Hello, world!"
    expected_hash = "315f5bdb76d078c43b8ac0064e4a0164612b1fce77c869345bfc94c75894edd3"
    assert _hash_str(test_string) == expected_hash


@unit_test
def test_convert_python_dtypes_to_panda_dtypes_error_unsupported_type() -> None:
    """Tests error is raised with unsupported type."""
    with pytest.raises(ValueError):
        _convert_python_dtypes_to_pandas_dtypes(set, "col_name")


@unit_test
def test_convert_python_dtypes_to_panda_dtypes_error_date_type() -> None:
    """Tests string is returned with date type."""
    dtype = _convert_python_dtypes_to_pandas_dtypes(datetime.date, "col_name")
    assert dtype == pd.StringDtype()


@unit_test
def test_convert_python_dtypes_to_panda_dtypes_error_datetime_type() -> None:
    """Tests string is returned with datetime type."""
    dtype = _convert_python_dtypes_to_pandas_dtypes(datetime.datetime, "col_name")
    assert dtype == pd.StringDtype()



"""Reusable fixtures for pytest tests."""
# Import all child modules' fixtures
from .authorisation_checker_fixtures import *  # noqa: F401, F403
from .encryption_fixtures import *  # noqa: F401, F403
from .env_fixtures import *  # noqa: F401, F403
from .helper_fixtures import *  # noqa: F401, F403
from .hub_and_am_fixtures import *  # noqa: F401, F403
from .schema_fixtures import *  # noqa: F401, F403
from .storage_fixtures import *  # noqa: F401, F403
from .transport_layer_fixtures import *  # noqa: F401, F403
from .web_utils_fixtures import *  # noqa: F401, F403



"""Tests for the federated transport utils."""
from bitfount.federated.transport.utils import _average_training_metrics
from tests.utils.helper import unit_test


@unit_test
def test_metrics_averaging() -> None:
    """Tests validation metrics are averaged and returned when provided."""
    validation_metrics = [{"AUC": "0.7", "F1": "0.8"}, {"AUC": "0.9", "F1": "0.6"}]

    metrics = _average_training_metrics(validation_metrics)
    assert metrics == {"AUC": 0.8, "F1": 0.7}


@unit_test
def test_metrics_averaging_no_metrics() -> None:
    """Tests aggregator returns no validation metrics when not provided."""
    metrics = _average_training_metrics([{}])
    assert metrics == {}



"""Tests for storage.py."""
import http
import json
from pathlib import Path
import pickle
import re
from typing import Dict, Final, Optional, Type, Union
from unittest.mock import AsyncMock, Mock

import httpx
import msgpack
import pytest
from pytest import fixture
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture
import requests
from requests import HTTPError, RequestException
from requests_toolbelt import MultipartDecoder
import responses
from responses.matchers import multipart_matcher

from bitfount import storage, web_utils
from bitfount.storage import (
    _DEFAULT_FILE_NAME,
    _async_download_data_from_s3,
    _async_download_from_s3,
    _async_upload_data_to_s3,
    _async_upload_to_s3,
    _download_data_from_s3,
    _download_file_from_s3,
    _download_from_s3,
    _get_packed_data_object_size,
    _upload_data_to_s3,
    _upload_file_to_s3,
    _upload_to_s3,
)
from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL
from tests.utils.helper import unit_test

_HTTP_STATUS_CODES: Final = tuple(i.value for i in http.HTTPStatus)


@fixture
def data() -> Dict[str, Union[int, str]]:
    """Data to upload."""
    return {"this": "is", "some": "data", "object": 1}


@fixture
def packed_data(data: Dict[str, Union[int, str]]) -> bytes:
    """Data form when it has been packed by msgpack."""
    return msgpack.dumps(json.dumps(data))


@fixture
def file_name() -> str:
    """Filename for data file."""
    return "data_file.txt"


@fixture
def custom_filename() -> str:
    """Custom filename to use for tests."""
    return "custom_filename.txt"


@fixture
def data_file(data: Dict[str, Union[int, str]], file_name: str, tmp_path: Path) -> Path:
    """Data written out as pickled binary to a file."""
    data_file_path = tmp_path / file_name
    with open(data_file_path, "wb") as f:
        pickle.dump(data, f)
    return data_file_path


@fixture
def file_data(data_file: Path) -> bytes:
    """Pickled data form when packed by msgpack."""
    with open(data_file, "rb") as f:
        return f.read()


@fixture
def bad_response_text() -> str:
    """Body text for a bad response."""
    return "BAD BODY"


@unit_test
class TestSynchronousS3Interactions:
    """Tests for the synchronous S3 interactions."""

    @responses.activate
    def test__upload_to_s3(
        self,
        file_name: str,
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _upload_to_s3 works correctly.

        Data should be uploaded with the expected fields.
        """
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields, files={"file": (file_name, packed_data)}
                )
            ],
        )

        _upload_to_s3(
            s3_upload_url, s3_upload_fields, to_upload=packed_data, file_name=file_name
        )

    @responses.activate
    def test__upload_to_s3_default_file_name(
        self,
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _upload_to_s3 works correctly.

        Data should be uploaded with the expected fields.
        """
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
        )

        _upload_to_s3(s3_upload_url, s3_upload_fields, to_upload=packed_data)

    @responses.activate
    def test__upload_to_s3_file_handler(
        self,
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        tmp_path: Path,
    ) -> None:
        """Tests _upload_to_s3 works correctly.

        Data should be uploaded with the expected fields.
        """
        with open(tmp_path / "packed_data", "wb") as f:
            f.write(packed_data)

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
        )

        with open(tmp_path / "packed_data", "rb") as f:
            _upload_to_s3(s3_upload_url, s3_upload_fields, to_upload=f)

    @responses.activate
    def test__upload_to_s3_failure_request_exception(
        self,
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _upload_to_s3 fails when RequestException encountered."""
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
            body=RequestException(),
        )

        # Error message
        with pytest.raises(RequestException, match="Issue uploading object to S3$"):
            _upload_to_s3(s3_upload_url, s3_upload_fields, packed_data)

    @pytest.mark.parametrize(
        "status_code",
        (400, 404, 500),
    )
    @responses.activate
    def test__upload_to_s3_failure_status_code(
        self,
        bad_response_text: str,
        packed_data: bytes,
        remove_web_retry_backoff_sleep: Optional[Mock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        status_code: int,
    ) -> None:
        """Tests _upload_to_s3 fails when status code is 400+."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
            status=status_code,
            body=bad_response_text,
        )

        with pytest.raises(
            HTTPError,
            match=re.escape(
                f"Issue uploading object to S3: ({status_code}) {bad_response_text}"
            ),
        ):
            _upload_to_s3(s3_upload_url, s3_upload_fields, packed_data)

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @responses.activate
    def test__upload_to_s3_failure_non_ok_status_code(
        self,
        bad_response_text: str,
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _upload_to_s3 fails when status code is not 200 or 201."""
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
            status=300,
            body=bad_response_text,
        )

        with pytest.raises(
            HTTPError,
            match=re.escape(f"Issue uploading object to S3: (300) {bad_response_text}"),
        ):
            _upload_to_s3(s3_upload_url, s3_upload_fields, packed_data)

    @responses.activate
    def test_upload_data_to_s3(
        self,
        data: Dict[str, Union[int, str]],
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_data_to_s3 works correctly.

        Data should be packed and uploaded with the expected fields.
        """
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, packed_data)},
                )
            ],
        )

        _upload_data_to_s3(s3_upload_url, s3_upload_fields, data=json.dumps(data))

    @responses.activate
    def test_upload_file_to_s3_with_path(
        self,
        data_file: Path,
        file_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with file path."""
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (data_file.name, file_data)},
                )
            ],
        )

        _upload_file_to_s3(s3_upload_url, s3_upload_fields, file_path=data_file)

    @responses.activate
    def test_upload_file_to_s3_with_path_and_file_name(
        self,
        custom_filename: str,
        data_file: Path,
        file_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with file path and custom filename."""
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (custom_filename, file_data)},
                )
            ],
        )

        _upload_file_to_s3(
            s3_upload_url,
            s3_upload_fields,
            file_path=data_file,
            file_name=custom_filename,
        )

    @responses.activate
    def test_upload_file_to_s3_with_str_file_contents(
        self,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with string file-contents."""
        str_contents: str = "Hello, world!"
        bytes_contents: bytes = str_contents.encode("utf-8")

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, bytes_contents)},
                )
            ],
        )

        _upload_file_to_s3(s3_upload_url, s3_upload_fields, file_contents=str_contents)

    @responses.activate
    def test_upload_file_to_s3_with_str_file_contents_and_custom_filename(
        self,
        custom_filename: str,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with string file-contents and custom filename."""  # noqa: B950
        str_contents: str = "Hello, world!"
        bytes_contents: bytes = str_contents.encode("utf-8")

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (custom_filename, bytes_contents)},
                )
            ],
        )

        _upload_file_to_s3(
            s3_upload_url,
            s3_upload_fields,
            file_contents=str_contents,
            file_name=custom_filename,
        )

    @responses.activate
    def test_upload_file_to_s3_with_str_file_contents_and_non_standard_encoding(
        self,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with string file-contents and encoding."""
        str_contents: str = "Hello, world!"
        encoding = "ascii"
        bytes_contents: bytes = str_contents.encode(encoding)

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, bytes_contents)},
                )
            ],
        )

        _upload_file_to_s3(
            s3_upload_url,
            s3_upload_fields,
            file_contents=str_contents,
            file_encoding=encoding,
        )

    @responses.activate
    def test_upload_file_to_s3_with_bytes_file_contents(
        self,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with bytes file-contents."""
        bytes_contents: bytes = b"Hello World!"

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (_DEFAULT_FILE_NAME, bytes_contents)},
                )
            ],
        )

        _upload_file_to_s3(
            s3_upload_url, s3_upload_fields, file_contents=bytes_contents
        )

    @responses.activate
    def test_upload_file_to_s3_with_bytes_file_contents_and_custom_filename(
        self,
        custom_filename: str,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 works with bytes file-contents and custom filename."""  # noqa: B950
        bytes_contents: bytes = b"Hello World!"

        responses.add(
            method=responses.POST,
            url=s3_upload_url,
            match=[
                multipart_matcher(
                    data=s3_upload_fields,
                    files={"file": (custom_filename, bytes_contents)},
                )
            ],
        )

        _upload_file_to_s3(
            s3_upload_url,
            s3_upload_fields,
            file_contents=bytes_contents,
            file_name=custom_filename,
        )

    @responses.activate
    def test_upload_file_to_s3_fails_if_contents_and_path_provided(
        self,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests upload_file_to_s3 fails if incompatible args."""
        with pytest.raises(
            ValueError,
            match=re.escape(
                "One of file_path and file_contents must be provided, but not both."
            ),
        ):
            _upload_file_to_s3(
                s3_upload_url, s3_upload_fields, file_path=Mock(), file_contents=Mock()
            )

    @responses.activate
    def test__download_from_s3(
        self, packed_data: bytes, s3_download_url: _S3PresignedURL
    ) -> None:
        """Tests _download_from_s3 works.

        Should download data, which will match the original data.
        """
        responses.add(
            method=responses.GET,
            url=s3_download_url,
            body=packed_data,
        )

        downloaded_data = _download_from_s3(s3_download_url)

        assert downloaded_data == packed_data

    @responses.activate
    def test__download_from_s3_fails_request_exception(
        self,
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Tests _download_from_s3 fails when RequestException encountered."""
        responses.add(
            method=responses.GET,
            url=s3_download_url,
            body=RequestException(),
        )

        with pytest.raises(
            RequestException, match="Issue whilst retrieving data from S3$"
        ):
            _download_from_s3(s3_download_url)

    @pytest.mark.parametrize("status_code", (400, 404, 500))
    @responses.activate
    def test__download_from_s3_fails_status_code_bad(
        self,
        bad_response_text: str,
        remove_web_retry_backoff_sleep: Optional[Mock],
        s3_download_url: _S3PresignedURL,
        status_code: int,
    ) -> None:
        """Tests _download_from_s3 fails when status code is 400+."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            method=responses.GET,
            url=s3_download_url,
            status=status_code,
            body=bad_response_text,
        )

        with pytest.raises(
            HTTPError,
            match=re.escape(
                f"Issue whilst retrieving data from S3:"
                f" ({status_code}) {bad_response_text}"
            ),
        ):
            _download_from_s3(s3_download_url)

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @responses.activate
    def test__download_from_s3_fails_status_code_not_good(
        self, bad_response_text: str, s3_download_url: _S3PresignedURL
    ) -> None:
        """Tests _download_from_s3 fails when status code is not 200/201."""
        responses.add(
            method=responses.GET,
            url=s3_download_url,
            status=202,
            body=bad_response_text,
        )

        with pytest.raises(
            HTTPError,
            match=re.escape(
                f"Issue whilst retrieving data from S3: (202) {bad_response_text}"
            ),
        ):
            _download_from_s3(s3_download_url)

    @responses.activate
    def test_download_data_from_s3(
        self,
        data: Dict[str, Union[int, str]],
        packed_data: bytes,
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Tests download_data_from_s3 works correctly.

        Data should be unpacked upon retrieval.
        """
        responses.add(
            method=responses.GET,
            url=s3_download_url,
            body=packed_data,
        )

        downloaded_data = _download_data_from_s3(s3_download_url)

        # Check was unpacked
        assert downloaded_data == json.dumps(data)

    @responses.activate
    def test_download_file_from_s3(
        self, packed_data: bytes, s3_download_url: _S3PresignedURL
    ) -> None:
        """Tests download_file_from_s3 works correctly.

        Data should be downloaded as is, and not encoded.
        """
        responses.add(
            method=responses.GET,
            url=s3_download_url,
            body=packed_data,
        )

        downloaded_data = _download_file_from_s3(s3_download_url)

        # Check was unpacked
        assert downloaded_data == packed_data

    @responses.activate
    def test_download_file_from_s3_with_encoding(
        self,
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Tests download_file_from_s3 works correctly with encoding.

        Retrieved data should be encoded as a string.
        """
        str_contents: str = "Hello, world!"
        bytes_contents: bytes = str_contents.encode("utf-8")

        responses.add(
            method=responses.GET,
            url=s3_download_url,
            body=bytes_contents,
        )

        downloaded_data = _download_file_from_s3(s3_download_url, encoding="utf-8")

        # Check was encoded correctly
        assert isinstance(downloaded_data, str)
        assert downloaded_data == str_contents

    def test_get_packed_data_object_size(
        self,
        data: Dict[str, Union[int, str]],
        packed_data: bytes,
    ) -> None:
        """Tests that get_packed_data_object_size works correctly."""
        expected_length = len(packed_data)

        assert _get_packed_data_object_size(json.dumps(data)) == expected_length

    @responses.activate
    def test_get_packed_data_object_size_corresponds_to_upload(
        self,
        data: Dict[str, Union[int, str]],
        packed_data: bytes,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests that get_packed_data_object_size matches upload size."""
        # Get output size and check it matches the packed data
        expected_length = _get_packed_data_object_size(json.dumps(data))
        assert len(packed_data) == expected_length

        # Do mocked upload to give us the request
        responses.add(
            method=responses.POST,
            url=s3_upload_url,
        )
        _upload_data_to_s3(s3_upload_url, s3_upload_fields, json.dumps(data))

        # Check request file size as actually uploaded
        request = responses.calls[0].request
        file_contents: Optional[bytes] = None
        # Extract file content bytes from the multipart/form-data request body
        mpd = MultipartDecoder(
            request.body, content_type=request.headers["content-type"]
        )
        # Find the file part of the body (will have "filename" in the disposition)
        for part in mpd.parts:
            content_disposition: bytes = part.headers[b"Content-Disposition"]
            if b"filename" in content_disposition:
                file_contents = part.content

        # Check that bytes length in request body matches expected_length
        assert file_contents is not None
        assert len(file_contents) == expected_length


@unit_test
class TestAsynchronousS3Interactions:
    """Tests for the asynchronous S3 interactions."""

    # -------------------------------------------------------------------------
    # TEST UPLOAD TO S3
    # -------------------------------------------------------------------------

    @pytest.mark.parametrize(
        "status_code", storage._OK_UPLOAD_STATUS_CODES, ids=lambda x: f"status_code={x}"
    )
    async def test_async_upload_to_s3(
        self,
        file_name: str,
        httpx_mock: HTTPXMock,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        status_code: int,
    ) -> None:
        """Tests _async_upload_to_s3 works correctly.

        Data should be uploaded with the expected fields.
        """
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        def custom_response(request: httpx.Request) -> httpx.Response:
            """Parse HTTP request, make assertions and return a response."""
            assert request.method == "POST"
            assert request.url == s3_upload_url
            assert "multipart/form-data" in request.headers["content-type"]
            assert file_name in request.read().decode("utf-8", errors="ignore")
            assert packed_data in request.content
            return httpx.Response(status_code=status_code)

        httpx_mock.add_callback(custom_response)

        await _async_upload_to_s3(
            s3_upload_url, s3_upload_fields, to_upload=packed_data, file_name=file_name
        )

    @pytest.mark.parametrize(
        "status_code",
        (s for s in _HTTP_STATUS_CODES if 400 <= s < 600),
        ids=lambda x: f"status_code={x}",
    )
    async def test_async_upload_to_s3_returns_bad_response(
        self,
        file_name: str,
        httpx_mock: HTTPXMock,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        status_code: int,
    ) -> None:
        """Tests _async_upload_to_s3 works correctly when the response is bad."""
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        def custom_response(request: httpx.Request) -> httpx.Response:
            """Parse HTTP request, make assertions and return a response."""
            assert request.method == "POST"
            assert request.url == s3_upload_url
            assert "multipart/form-data" in request.headers["content-type"]
            assert file_name in request.read().decode("utf-8", errors="ignore")
            assert packed_data in request.content
            return httpx.Response(status_code=status_code)

        httpx_mock.add_callback(custom_response)

        with pytest.raises(
            HTTPError, match=re.escape(f"Issue uploading object to S3: ({status_code})")
        ):
            await _async_upload_to_s3(
                s3_upload_url,
                s3_upload_fields,
                to_upload=packed_data,
                file_name=file_name,
            )

        # If we expected automatic retry to be triggered, check it was
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_async_web_retry_backoff_sleep.await_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            assert remove_async_web_retry_backoff_sleep.await_count == 0

    @pytest.mark.parametrize(
        "status_code",
        (
            s
            for s in _HTTP_STATUS_CODES
            if 100 <= s < 400 and s not in storage._OK_UPLOAD_STATUS_CODES
        ),
        ids=lambda x: f"status_code={x}",
    )
    async def test_async_upload_to_s3_returns_unacceptable_response(
        self,
        file_name: str,
        httpx_mock: HTTPXMock,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
        status_code: int,
    ) -> None:
        """Tests _async_upload_to_s3 works correctly when the response is unacceptable.

        The response here is not necessarily bad such that it won't raise an exception
        when `raise_for_status` is called but has been defined as not ok by ourselves.
        """
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_response(status_code=status_code)

        with pytest.raises(
            HTTPError, match=re.escape(f"Issue uploading object to S3: ({status_code})")
        ):
            await _async_upload_to_s3(
                s3_upload_url,
                s3_upload_fields,
                to_upload=packed_data,
                file_name=file_name,
            )

    @pytest.mark.parametrize(
        argnames=("exc_to_raise", "expected_exc_cls"),
        argvalues=(
            pytest.param(
                httpx.HTTPStatusError("TEST", request=Mock(), response=Mock()),
                requests.HTTPError,
                id="httpx.HTTPStatusError",
            ),
            pytest.param(
                httpx.ConnectError("TEST", request=Mock()),
                requests.ConnectionError,
                id="httpx.ConnectError",
            ),
            pytest.param(
                httpx.ReadTimeout("TEST", request=Mock()),
                requests.RequestException,
                id="httpx.ReadTimeout",
            ),
            # This exception results in different handling
            pytest.param(
                httpx.ConnectTimeout("TEST", request=Mock()),
                requests.RequestException,
                id="httpx.ConnectTimeout",
            ),
        ),
    )
    async def test_async_upload_to_s3_raises_exception(
        self,
        exc_to_raise: httpx.HTTPError,
        expected_exc_cls: Type[requests.RequestException],
        file_name: str,
        httpx_mock: HTTPXMock,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _async_upload_to_s3 handles exception when request is made.

        Tests that the exception is handled and re-raised.
        """
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_exception(exc_to_raise)

        with pytest.raises(
            expected_exc_cls, match="Issue uploading object to S3.*TEST"
        ):
            await _async_upload_to_s3(
                s3_upload_url,
                s3_upload_fields,
                to_upload=packed_data,
                file_name=file_name,
            )

        # Check expected retry
        if isinstance(exc_to_raise, httpx.ConnectError):
            assert (
                remove_async_web_retry_backoff_sleep.await_count
                == web_utils._DEFAULT_MAX_RETRIES
            )

    async def test_async_upload_data_to_s3(
        self,
        data: Dict[str, Union[int, str]],
        mocker: MockerFixture,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Tests _async_download_data_to_s3 works correctly."""
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        mock_async_upload = mocker.patch("bitfount.storage._async_upload_to_s3")
        await _async_upload_data_to_s3(
            s3_upload_url, s3_upload_fields, json.dumps(data)
        )
        mock_async_upload.assert_awaited_once_with(
            s3_upload_url, s3_upload_fields, packed_data
        )

    # -------------------------------------------------------------------------
    # TEST DOWNLOAD FROM S3
    # -------------------------------------------------------------------------

    @pytest.mark.parametrize(
        "status_code",
        storage._OK_DOWNLOAD_STATUS_CODES,
        ids=lambda x: f"status_code={x}",
    )
    async def test_async_download_from_s3(
        self,
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_download_url: _S3PresignedURL,
        status_code: int,
    ) -> None:
        """Tests _async_download_from_s3 works correctly."""
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        def custom_response(request: httpx.Request) -> httpx.Response:
            """Parse HTTP request, make assertions and return a response."""
            assert request.method == "GET"
            assert request.url == s3_download_url
            return httpx.Response(status_code=status_code)

        httpx_mock.add_callback(custom_response)

        await _async_download_from_s3(s3_download_url)

    @pytest.mark.parametrize(
        "status_code",
        (s for s in _HTTP_STATUS_CODES if 400 <= s < 600),
        ids=lambda x: f"status_code={x}",
    )
    async def test_async_download_from_s3_returns_bad_response(
        self,
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_download_url: _S3PresignedURL,
        status_code: int,
    ) -> None:
        """Tests _async_download_from_s3 works correctly when the response is bad."""
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        def custom_response(request: httpx.Request) -> httpx.Response:
            """Parse HTTP request, make assertions and return a response."""
            assert request.method == "GET"
            assert request.url == s3_download_url
            return httpx.Response(status_code=status_code)

        httpx_mock.add_callback(custom_response)

        with pytest.raises(
            HTTPError,
            match=re.escape(f"Issue whilst retrieving data from S3: ({status_code})"),
        ):
            await _async_download_from_s3(s3_download_url)

    @pytest.mark.parametrize(
        "status_code",
        (
            s
            for s in _HTTP_STATUS_CODES
            if 100 <= s < 400 and s not in storage._OK_DOWNLOAD_STATUS_CODES
        ),
        ids=lambda x: f"status_code={x}",
    )
    async def test_async_download_from_s3_returns_unacceptable_response(
        self,
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_download_url: _S3PresignedURL,
        status_code: int,
    ) -> None:
        """Tests _async_download_from_s3 works when the response is unacceptable.

        The response here is not necessarily bad such that it won't raise an exception
        when `raise_for_status` is called but has been defined as not ok by ourselves.
        """
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_response(status_code=status_code)

        with pytest.raises(
            HTTPError,
            match=re.escape(f"Issue whilst retrieving data from S3: ({status_code})"),
        ):
            await _async_download_from_s3(s3_download_url)

    @pytest.mark.parametrize(
        argnames=("exc_to_raise", "expected_exc_cls"),
        argvalues=(
            pytest.param(
                httpx.HTTPStatusError("TEST", request=Mock(), response=Mock()),
                requests.HTTPError,
                id="httpx.HTTPStatusError",
            ),
            pytest.param(
                httpx.ConnectError("TEST", request=Mock()),
                requests.ConnectionError,
                id="httpx.ConnectError",
            ),
            pytest.param(
                httpx.ReadTimeout("TEST", request=Mock()),
                requests.RequestException,
                id="httpx.ReadTimeout",
            ),
            # This exception results in different handling
            pytest.param(
                httpx.ConnectTimeout("TEST", request=Mock()),
                requests.RequestException,
                id="httpx.ConnectTimeout",
            ),
        ),
    )
    async def test_async_download_from_s3_raises_exception(
        self,
        exc_to_raise: httpx.HTTPError,
        expected_exc_cls: Type[requests.RequestException],
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Tests _async_download_from_s3 handles exception when request is made.

        Tests that the exception is handled and re-raised.
        """
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_exception(exc_to_raise)

        with pytest.raises(
            expected_exc_cls, match="Issue whilst retrieving data from S3.*TEST"
        ):
            await _async_download_from_s3(s3_download_url)

        # Check expected retry
        if isinstance(exc_to_raise, httpx.ConnectError):
            assert (
                remove_async_web_retry_backoff_sleep.await_count
                == web_utils._DEFAULT_MAX_RETRIES
            )

    async def test_async_download_data_from_s3(
        self,
        data: Dict[str, Union[int, str]],
        mocker: MockerFixture,
        packed_data: bytes,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Tests _async_download_data_from_s3 works correctly."""
        # Check that backoff sleep has been correctly mocked
        assert remove_async_web_retry_backoff_sleep is not None

        mock_async_download = mocker.patch(
            "bitfount.storage._async_download_from_s3", return_value=packed_data
        )
        result = await _async_download_data_from_s3(s3_download_url)
        assert result == json.dumps(data)
        mock_async_download.assert_awaited_once_with(s3_download_url)



"""Module for handling logging messages out of a Jupyter Notebook."""
import errno
import inspect
import logging
from logging import Logger, LogRecord
from multiprocessing.managers import BaseManager
from queue import Empty, Queue
import random
import secrets
import socket
import textwrap
from threading import Event, Thread
from typing import Final, Generator, List

from pytest import fixture

logger = logging.getLogger(__name__)


def _get_random_port() -> int:
    # Max of 10 attempts
    for _ in range(10):
        # Recommended ephemeral port range from IANA
        port = random.randint(49152, 65535)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                # Attempt to bind to the port. If successful, then return the port
                # which will close the socket immediately, freeing the port up to be
                # used elsewhere
                s.bind(("", port))
                return port
            except OSError as e:
                if e.errno in (errno.EADDRINUSE, errno.EADDRNOTAVAIL):
                    # Address already unavailable so try again
                    continue
                else:
                    raise e

    raise RuntimeError("Unable to get random port for log queue")


_LOG_QUEUE_PORT: Final[int] = _get_random_port()
_LOG_QUEUE_AUTH_KEY: Final[bytes] = secrets.token_hex(10).encode()
_RECORD_GET_TIMEOUT: Final[int] = 1

# Shared queue for logging messages to and from
_LOG_QUEUE: Final[Queue] = Queue()


def _get_log_queue() -> Queue:
    """Return the shared queue for logging messages."""
    return _LOG_QUEUE


class LogQueueManager(BaseManager):
    """Custom multiprocessing Manager for handling log messages."""

    pass


LogQueueManager.register("get_log_queue", callable=_get_log_queue)


def _log_retriever(log_queue: Queue, stopper: Event) -> None:
    """Listen to queue, log received records until stop is set."""
    logger.info("Log retriever started")
    while not stopper.is_set():
        try:
            record: LogRecord = log_queue.get(timeout=_RECORD_GET_TIMEOUT)
            # Update log info with annotations
            notebook_name: str = record.notebook  # type: ignore[attr-defined] # Reason: we ensure this is added using LogQueueFilter below # noqa: B950
            record.processName = f"[{notebook_name}_notebook]"
            record.msg = f"[FROM {notebook_name.upper()} NOTEBOOK] {record.msg}"
            # Send to handler
            logger.handle(record)
        except Empty:
            pass
    logger.info("Log retriever stopped")


@fixture(scope="module")
def log_queue_manager() -> Generator[LogQueueManager, None, None]:
    """Hosts and starts a LogQueueManager instance for logging messages."""
    with LogQueueManager(
        address=("", _LOG_QUEUE_PORT), authkey=_LOG_QUEUE_AUTH_KEY
    ) as manager:
        logger.info("LogQueueManager started")
        yield manager


@fixture
def clear_log_queue() -> None:
    """Clears the log queue, ready for the next test."""
    # Clear queue, using direct reference as otherwise will hit Manager's proxy
    logger.info("Clearing log queue")
    with _LOG_QUEUE.mutex:
        _LOG_QUEUE.queue.clear()
    logger.info("Log queue cleared")


@fixture(scope="module", autouse=True)
def log_queue_listener(
    log_queue_manager: LogQueueManager,
) -> Generator[None, None, None]:
    """Function-scoped listener to retrieve and log from the LogQueueManager."""
    # Get shared proxy of queue and create thread to listen to it
    log_queue: Queue = log_queue_manager.get_log_queue()  # type: ignore[attr-defined] # Reason: registered on Manager above # noqa: B950
    stop_event: Event = Event()
    log_retrieval_thread = Thread(
        target=_log_retriever,
        name="log retriever",
        args=(log_queue, stop_event),
        daemon=True,
    )
    logger.info("Starting log retrieval thread")
    log_retrieval_thread.start()

    yield

    logger.info("Stopping log retrieval thread")
    stop_event.set()
    log_retrieval_thread.join(_RECORD_GET_TIMEOUT + 1)


# Just a placeholder to mimic it being in scope in the notebook
loggers: Final[List[Logger]] = []
_NOTEBOOK_PLACEHOLDER: Final[str] = "_NOTEBOOK_PLACEHOLDER"


def _add_notebook_queue_logging() -> None:
    """Code to be injected into notebooks to setup queue logging."""
    import logging
    from logging import Filter, LogRecord
    from logging.handlers import QueueHandler
    from multiprocessing.managers import BaseManager

    # Create copy of LogQueueManager class to have one in scope of injected code
    class LogQueueManager(BaseManager):
        pass

    LogQueueManager.register("get_log_queue")

    # Connect to Manager and get shared log queue
    manager = LogQueueManager(
        address=("", _LOG_QUEUE_PORT), authkey=_LOG_QUEUE_AUTH_KEY
    )  # needs to be inlined because of how it's injected
    manager.connect()
    log_queue = manager.get_log_queue()  # type: ignore[attr-defined] # Reason: registered on Manager above # noqa: B950

    # Create queue handler for loggers
    queue_handler = QueueHandler(log_queue)
    queue_handler.setLevel("INFO")

    # Create custom filter to append additional LogRecord field
    class LogQueueFilter(Filter):
        def filter(self, record: LogRecord) -> bool:
            record.notebook = _NOTEBOOK_PLACEHOLDER
            return True

    queue_handler.addFilter(LogQueueFilter())

    # Attach to root logger and set others to propagate
    logging.info("Attaching queue handler to root logger")
    logging.root.addHandler(queue_handler)
    for i_logger in loggers:
        logging.info(f"Setting {i_logger.name} to propagate")
        i_logger.propagate = True


def notebook_queue_logging_code(notebook_name: str) -> str:
    """Generates code to inject into a notebook to allow queue-logging.

    Args:
        notebook_name: The name to use to annotate log messages from the notebook.

    Returns:
        The generated code to inject.
    """
    source = inspect.getsource(_add_notebook_queue_logging)

    # Replace constants that aren't in scope
    source = source.replace("_LOG_QUEUE_PORT", str(_LOG_QUEUE_PORT))
    source = source.replace(
        "_LOG_QUEUE_AUTH_KEY", str(_LOG_QUEUE_AUTH_KEY)
    )  # will include quote marks
    source = source.replace("_NOTEBOOK_PLACEHOLDER", f'"{notebook_name}"')

    return textwrap.dedent(source) + f"{_add_notebook_queue_logging.__name__}()"



"""Contains the base classes for handling custom models."""

from abc import ABC, abstractmethod
import os
from typing import Any, List, Optional, Union

import isort

from bitfount.models.base_models import _BaseModel
from bitfount.utils import _get_object_source_code, delegates


# This class must implement `DistributedModelProtocol` but cannot inherit from it for
# two reasons:
# 1. Pytorch lightning does not like this for some reason and throws an error in the
# `PyTorchBitfountModel` subclass:
#   "AttributeError: cannot assign module before Module.__init__() call"
# 2. If this inherits from `DistributedModelProtocol`, mypy does not ensure that an
# implementation actually implements the protocol. As a result, we have to do an
# `isinstance` check which will always return `True` if the protocol is part of the
# hierarchy regardless of whether the implementation actually implements the protocol.
@delegates()
class BitfountModel(_BaseModel, ABC):
    """Base class for custom models which must implement `DistributedModelProtocol`.

    A base tagging class to highlight custom models which are designed to be uploaded to
    Bitfount Hub.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.class_name = type(self).__name__  # overrides _BaseModel attribute

    @staticmethod
    @abstractmethod
    def _get_import_statements() -> List[str]:
        """Returns a list of import statements likely to be required for the model.

        These are likely to be backend-specific so this method is left to be implemented
        by the subclass.

        Returns:
            A list of import statements.
        """
        raise NotImplementedError

    @classmethod
    def serialize_model_source_code(
        cls,
        filename: Union[str, os.PathLike],
        extra_imports: Optional[List[str]] = None,
    ) -> None:
        """Serializes the source code of the model to file.

        This is required so that the model source code can be uploaded to Bitfount Hub.

        Args:
            filename: The filename to save the source code to.
            extra_imports: A list of extra import statements to include in the source
                code.
        """
        source = _get_object_source_code(cls)
        all_imports = cls._get_import_statements()
        if extra_imports is not None:
            all_imports += extra_imports
        with open(filename, "w") as f:
            for import_ in all_imports:
                f.write(import_ + "\n")
            f.write(source)

        # Remove unused imports. It is recommended that `isort` and `black` be used
        # to format the source code after `autoimport` is used.
        # Security warnings ignored because these are only run on the user's own machine
        os.system(  # nosec[start_process_with_a_shell] # Reason: See above
            f"autoimport {str(filename)} > /dev/null 2>&1"  # Suppress stdout and stderr
        )

        # Sort the remaining imports
        isort.file(str(filename))

        # Format the source code
        # Security warnings ignored because these are only run on the user's own machine
        os.system(  # nosec[start_process_with_a_shell] # Reason: See above
            f"black {str(filename)} > /dev/null 2>&1"  # Suppress stdout and stderr
        )



"""Tests for the transformations reference extractions."""
from typing import Optional

import pytest
from pytest import fixture

from bitfount.transformations.exceptions import (
    IncorrectReferenceError,
    NotColumnReferenceError,
    NotTransformationReferenceError,
)
from bitfount.transformations.references import (
    _extract_col_ref,
    _extract_ref,
    _extract_transformation_ref,
)
from tests.utils.helper import unit_test

ref_name: str = "coltran"  # value allows testing we don't catch non-prefix part


def base_ref(prefix: Optional[str] = None) -> str:
    """Returns a base reference to build from."""
    if prefix:
        return f"{prefix}:{ref_name}"
    else:
        return ref_name


@fixture(scope="session")
def good_t_ref_short() -> str:
    """Returns a short-form transformation reference."""
    return base_ref("t")


@fixture(scope="session")
def good_t_ref_mid() -> str:
    """Returns a mid-form transformation reference."""
    return base_ref("tran")


@fixture(scope="session")
def good_t_ref_long() -> str:
    """Returns a long-form transformation reference."""
    return base_ref("transformation")


@fixture(scope="session")
def good_c_ref_short() -> str:
    """Returns a short-form column reference."""
    return base_ref("c")


@fixture(scope="session")
def good_c_ref_mid() -> str:
    """Returns a mid-form column reference."""
    return base_ref("col")


@fixture(scope="session")
def good_c_ref_long() -> str:
    """Returns a long-form column reference."""
    return base_ref("column")


@fixture(scope="session")
def bad_ref() -> str:
    """Returns a bad (i.e. unprefixed) reference."""
    return base_ref()


@unit_test
class TestExtractTransformationRef:
    """Tests for extract_transformation_ref()."""

    def test_extract_transformation_ref(
        self, good_t_ref_long: str, good_t_ref_mid: str, good_t_ref_short: str
    ) -> None:
        """Tests that extract_transformation_ref() works correctly."""
        assert ref_name == _extract_transformation_ref(good_t_ref_short)
        assert ref_name == _extract_transformation_ref(good_t_ref_mid)
        assert ref_name == _extract_transformation_ref(good_t_ref_long)

    def test_extract_transformation_ref_fails_no_prefix(self, bad_ref: str) -> None:
        """Tests that extract_transformation_ref() fails when there is no prefix."""
        with pytest.raises(
            NotTransformationReferenceError,
            match="Incorrect format for transformation reference",
        ):
            _extract_transformation_ref(bad_ref)

    def test_extract_transformation_ref_fails_wrong_type(self) -> None:
        """Tests that extract_transformation_ref() fails when ref non-string."""
        with pytest.raises(
            NotTransformationReferenceError,
            match="Incorrect type for transformation reference",
        ):
            _extract_transformation_ref(0)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950


@unit_test
class TestExtractColRef:
    """Tests for extract_col_ref()."""

    def test_extract_col_ref(
        self, good_c_ref_long: str, good_c_ref_mid: str, good_c_ref_short: str
    ) -> None:
        """Tests that extract_col_ref() works correctly."""
        assert ref_name == _extract_col_ref(good_c_ref_short)
        assert ref_name == _extract_col_ref(good_c_ref_mid)
        assert ref_name == _extract_col_ref(good_c_ref_long)

    def test_extract_col_ref_fails_no_prefix(self, bad_ref: str) -> None:
        """Tests that extract_col_ref() fails when there is no prefix."""
        with pytest.raises(
            NotColumnReferenceError, match="Incorrect format for column reference"
        ):
            _extract_col_ref(bad_ref)

    def test_extract_col_ref_fails_wrong_type(self) -> None:
        """Tests that extract_col_ref() fails when ref non-string."""
        with pytest.raises(
            NotColumnReferenceError, match="Incorrect type for column reference"
        ):
            _extract_col_ref(0)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950


@unit_test
class TestExtractRef:
    """Tests for _extract_ref()."""

    def test__extract_ref_transformation_ref(
        self, good_t_ref_long: str, good_t_ref_mid: str, good_t_ref_short: str
    ) -> None:
        """Tests that _extract_ref() works with transformations references."""
        assert ref_name == _extract_ref(good_t_ref_short)
        assert ref_name == _extract_ref(good_t_ref_mid)
        assert ref_name == _extract_ref(good_t_ref_long)

    def test__extract_ref_col_ref(
        self, good_c_ref_long: str, good_c_ref_mid: str, good_c_ref_short: str
    ) -> None:
        """Tests that _extract_ref() works with column references."""
        assert ref_name == _extract_ref(good_c_ref_short)
        assert ref_name == _extract_ref(good_c_ref_mid)
        assert ref_name == _extract_ref(good_c_ref_long)

    def test__extract_ref_fails_no_prefix(self, bad_ref: str) -> None:
        """Tests that _extract_ref() fails when there is no prefix."""
        with pytest.raises(
            IncorrectReferenceError,
            match="Argument is not a transformation or column reference.",
        ):
            _extract_ref(bad_ref)

    def test_extract_col_ref_fails_wrong_type(self) -> None:
        """Tests that _extract_ref() fails when ref non-string."""
        with pytest.raises(
            IncorrectReferenceError,
            match="Argument is not a transformation or column reference.",
        ):
            _extract_ref(0)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950



"""Defines functions which compute metrics."""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from functools import partial
import logging
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
    cast,
)

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import kstest
from sklearn.metrics import (
    accuracy_score,
    brier_score_loss,
    confusion_matrix,
    f1_score,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
)

from bitfount.models.base_models import ClassifierMixIn, RegressorMixIn, _BaseModel
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.types import DistributedModelProtocol

__all__: List[str] = [
    "BINARY_CLASSIFICATION_METRICS",
    "ClassificationMetric",
    "MULTICLASS_CLASSIFICATION_METRICS",
    "MULTILABEL_CLASSIFICATION_METRICS",
    "Metric",
    "MetricCollection",
    "MetricsProblem",
    "REGRESSION_METRICS",
    "SEGMENTATION_METRICS",
    "sum_fpr_curve",
]

_logger = logging.getLogger(__name__)


def _binary_dice_coefficient(target: np.ndarray, pred: np.ndarray) -> float:
    """Computes the dice coefficient for a binary segmentation task.

    Dice Coefficient is 2 * the Area of Overlap divided by the total
    number of pixels in both images.

    Args:
        pred: The predicted segmentation.
        target: The target segmentation.

    Returns:
        The binary dice coefficients score.
    """
    y_true_f = target.flatten()
    y_pred_f = pred.flatten()
    intersection = (y_true_f * y_pred_f).sum()
    smooth = 0.0001
    return cast(
        float, (2.0 * intersection) / (y_true_f.sum() + y_pred_f.sum() + smooth)
    )


def multiclass_dice_coefficients(target: np.ndarray, pred: np.ndarray) -> float:
    """Computes the dice coefficient for a multiclass segmentation task.

    Args:
        pred: The predicted segmentation.
        target: The target segmentation of shape:
            (Batch_size X number of classes X Height X Weight).

    Returns:
        The multiclass dice coefficients score.
    """
    n_classes = pred.shape[1]
    dice = 0.0
    for index in range(1, n_classes):
        dice += _binary_dice_coefficient(target, pred[:, index, :, :])

    return dice / n_classes  # taking average


def _stats_score(
    target: np.ndarray, pred: np.ndarray, n: int = 1
) -> Tuple[int, int, int, int]:
    """Computes the relevant statistics between targets and predictions.

    More specifically, computes the true positive (tp), false positive(fp),
    true negative (tn) and false positive (fp) rates given the targets,
    prediction and the class.

    Args:
        pred: The predicted value.
        target: The target value.

    Returns:
        A tuple containing (tp, fp, tn, fn).
    """
    cls_pred = np.argmax(pred, axis=1)
    tp = ((cls_pred == n) * (target == n)).sum()
    fp = ((cls_pred == n) * (target != n)).sum()
    tn = ((cls_pred != n) * (target != n)).sum()
    fn = ((cls_pred != n) * (target == n)).sum()
    return tp, fp, tn, fn


def dice_score(target: np.ndarray, pred: np.ndarray) -> float:
    """Computes the dice score.

    This is computed as 2*tp/2*tp+fp+fn.

    Args:
        pred: The predicted segmentation.
        target: The target segmentation.

    Returns:
        The dice score.
    """
    n_classes = pred.shape[1]
    score = 0.0
    # only loop from 1 as we don't care about the background
    for i in range(1, n_classes):
        tp, fp, tn, fn = _stats_score(target, pred, i)
        denom = 2 * tp + fp + fn
        if not (target == i).any():
            # no foreground class
            score_cls = 0.0
        elif np.isclose(denom, np.zeros_like(denom)).any():
            # nan result
            score_cls = 0.0
        else:
            score_cls = (2 * tp) / denom
        score += score_cls
    return score / (n_classes)


def iou(target: np.ndarray, pred: np.ndarray) -> float:
    """Computes the Intersection over union score.

    This is computed as tp/tp+fp+fn. Also known as Jaccard Index.

    Args:
        pred: The predicted segmentation.
        target: The target segmentation.

    Returns:
        The IoU score.
    """
    # computed as tp/tp+fp+fn
    n_classes = pred.shape[1]
    score = 0.0
    # only loop from 1 as we don't care about the background
    for i in range(1, n_classes):
        tp, fp, tn, fn = _stats_score(target, pred, i)
        denom = tp + fp + fn
        if not (target == i).any():
            # no foreground class
            score_cls = 0.0
        elif np.isclose(denom, np.zeros_like(denom)).any():
            # nan result
            score_cls = 0.0
        else:
            score_cls = (tp) / denom
        score += score_cls
    return score / (n_classes)


@dataclass
class Metric:
    """A metric used for assessing ML model performance.

    Attributes:
        func: A function which computes the metric. Must take two arguments:
            y_true and y_pred and return a metric as a float.
    """

    func: Callable[[np.ndarray, np.ndarray], float]


@delegates()
@dataclass
class ClassificationMetric(Metric):
    """A classification metric used for assessing ML model performance.

    Args:
        probabilities: Whether y_pred needs to be classes or probabilities.
    """

    probabilities: bool


MULTICLASS_CLASSIFICATION_METRICS: Dict[str, Metric] = {
    "AUC": ClassificationMetric(
        partial(roc_auc_score, multi_class="ovr", average="macro"), True
    ),
    "F1": ClassificationMetric(partial(f1_score, average="macro"), False),
    "Recall": ClassificationMetric(partial(recall_score, average="macro"), False),
    "Precision": ClassificationMetric(
        partial(precision_score, average="macro", zero_division=0), False
    ),
    "Accuracy": ClassificationMetric(accuracy_score, False),
}

MULTILABEL_CLASSIFICATION_METRICS: Dict[str, Metric] = {
    "AUC": ClassificationMetric(
        partial(roc_auc_score, multi_class="ovr", average="micro"), True
    ),
    "F1": ClassificationMetric(partial(f1_score, average="micro"), False),
    "Recall": ClassificationMetric(partial(recall_score, average="micro"), False),
    "Precision": ClassificationMetric(
        partial(precision_score, average="micro", zero_division=0), False
    ),
    "Accuracy": ClassificationMetric(accuracy_score, False),
}

BINARY_CLASSIFICATION_METRICS: Dict[str, Metric] = {
    "AUC": ClassificationMetric(
        partial(roc_auc_score, multi_class="ovr", average="macro"), True
    ),
    "F1": ClassificationMetric(partial(f1_score, average="binary"), False),
    "Recall": ClassificationMetric(partial(recall_score, average="binary"), False),
    "Precision": ClassificationMetric(
        partial(precision_score, average="binary", zero_division=0), False
    ),
    "Accuracy": ClassificationMetric(accuracy_score, False),
    "BrierLoss": ClassificationMetric(brier_score_loss, True),
}

_rmse_func: Callable = lambda x, y: np.sqrt(mean_squared_error(x, y))

REGRESSION_METRICS: Dict[str, Metric] = {
    "MAE": Metric(mean_absolute_error),
    "MSE": Metric(mean_squared_error),
    "RMSE": Metric(_rmse_func),
    "R2": Metric(r2_score),
    "KS": Metric(kstest),
}

SEGMENTATION_METRICS: Dict[str, Metric] = {
    "IoU": Metric(iou),
    "DiceCoefficients": Metric(multiclass_dice_coefficients),
    "DiceScore": Metric(dice_score),
}


class MetricsProblem(Enum):
    """Simple wrapper for different problem types for MetricCollection."""

    BINARY_CLASSIFICATION = auto()
    MULTICLASS_CLASSIFICATION = auto()
    MULTILABEL_CLASSIFICATION = auto()
    REGRESSION = auto()
    SEGMENTATION = auto()


class MetricCollection:
    """Container class for metrics to calculate.

    Args:
        metrics: A list of metrics to calculate.
        problem: The problem type. If metrics are not specified, the problem type will
            be used to determine the metrics to calculate.

    Raises:
        ValueError: If neither one of `problem` nor `metrics` is specified.

    Attributes:
        metrics: A list of metrics to calculate.
        problem: The problem type.
        optimal_threshold: The optimal threshold to separate classes (only used for
            classification problems).
        thresholds: The thresholds to separate classes (only used for
            classification problems).
        threshold_metrics: The metrics for each threshold (only used for
            classification problems).
    """

    def __init__(
        self,
        metrics: Optional[MutableMapping[str, Metric]] = None,
        problem: Optional[MetricsProblem] = None,
    ):
        if metrics is None and problem is None:
            raise ValueError(
                "At least one of `metrics` or `problem` must be specified."
            )

        if metrics is not None:
            self.metrics = metrics
        else:
            self.metrics = self._get_metrics(cast(MetricsProblem, problem))

        self.problem = problem
        self.optimal_threshold: Optional[float] = None
        self.results: Dict[str, float] = {}
        self.thresholds = np.linspace(0, 100, 201, endpoint=True) / 100
        self.threshold_metrics: Optional[Mapping[str, Iterable[float]]] = None

    @classmethod
    def create_from_model(
        cls,
        model: Union[_BaseModel, DistributedModelProtocol],
        metrics: Optional[MutableMapping[str, Metric]] = None,
    ) -> MetricCollection:
        """Creates a MetricCollection object from a _BaseModel.

        Args:
            model: A _BaseModel.
            metrics: The metrics dictionary. Defaults to None.

        Returns:
            Instance of MetricCollection.
        """
        problem: Optional[MetricsProblem] = None
        # The casts are to reassure mypy that a subclass of both _BaseModel and
        # ClassifierMixIn can exist (the ._Schema attribute causes problems)
        if isinstance(cast(ClassifierMixIn, model), ClassifierMixIn):
            classifier_model = cast(ClassifierMixIn, model)
            if classifier_model.multilabel:
                problem = MetricsProblem.MULTILABEL_CLASSIFICATION
            elif classifier_model.n_classes > 2:
                problem = MetricsProblem.MULTICLASS_CLASSIFICATION
            else:
                problem = MetricsProblem.BINARY_CLASSIFICATION
        elif isinstance(cast(RegressorMixIn, model), RegressorMixIn):
            problem = MetricsProblem.REGRESSION
        else:
            _logger.warning("Metrics problem type can't be determined. Leaving empty.")

        return cls(problem=problem, metrics=metrics)

    def compute(
        self,
        test_target: np.ndarray,
        test_preds: np.ndarray,
        metric_to_optimise: str = "F1",
        threshold: Optional[float] = None,
    ) -> Dict[str, float]:
        """Compute list of metrics and save results in self.results.

        :::note

        Thresholds do not apply to multiclass problems.

        :::

        Args:
            test_target: A list of targets.
            test_preds: A list of predictions.
            metric_to_optimise: What metric to optimize in order to compute the
                optimal threshold. This will have no effect if there aren't any metrics
                to which a threshold is applied. Must be present in 'self.metrics'.
            threshold: If this argument is provided, this threshold will be
                used instead of optimising the threshold as per 'optimise'
        """
        if (
            self.problem is MetricsProblem.BINARY_CLASSIFICATION
            and np.asarray(test_preds).ndim == 2
        ):
            test_preds = np.asarray([row[1] for row in test_preds])

        if (
            self.problem is MetricsProblem.REGRESSION
            or self.problem is MetricsProblem.BINARY_CLASSIFICATION
        ):
            test_preds = np.asarray(test_preds).squeeze()
            test_target = np.asarray(test_target).squeeze()

        for metric_name, metric in self.metrics.items():
            if isinstance(metric, ClassificationMetric) and not metric.probabilities:
                test_probs_or_classes = self._get_prediction_classes(
                    test_target, test_preds, metric_to_optimise, threshold
                )
            else:
                test_probs_or_classes = test_preds
            try:
                self.results[metric_name] = cast(
                    float,
                    self._round(metric.func(test_target, test_probs_or_classes), 4),
                )
                # metric.func always returns a float, so the _round
                # function will return a float as well
            except (ValueError, IndexError):
                # This usually occurs when the targets do not have all possible classes
                self.results[metric_name] = np.inf
                _logger.warning(
                    "Unable to compute metric: " + metric_name + " setting to np.inf"
                )

        return self.results

    @staticmethod
    def _round(
        metric: Union[float, Tuple[float]], num_decimal_places: int
    ) -> Union[float, Tuple[float, ...]]:
        """Simply rounds metric by num_decimal_places and returns it."""
        if isinstance(metric, tuple):
            rounded_metric = tuple(round(i, num_decimal_places) for i in metric)
        elif isinstance(metric, float):
            rounded_metric = round(metric, num_decimal_places)  # type: ignore[arg-type] # reason: see below # noqa: B950
            # mypy raises the following error: Argument 1 to
            # "round" has incompatible type "float".

        return rounded_metric

    def get_results_df(self) -> pd.DataFrame:
        """Calls get_results and returns results as a dataframe."""
        return pd.DataFrame.from_dict(
            cast(Dict, self.results), orient="index", columns=["value"]
        )

    def view_thresholds(self) -> matplotlib.figure.Figure:
        """Displays chart of how thresholds affect relevant metrics."""
        fig, ax = plt.subplots()

        for metric_name, metric_vals in cast(
            Mapping[str, Iterable[float]], self.threshold_metrics
        ).items():
            plt.plot(self.thresholds, metric_vals, label=metric_name)
        ax.legend()
        ax.set_title("Metric values at various thresholds")
        ax.set_ylabel("Metric value")
        ax.set_xlabel("Threshold")
        plt.close()

        return fig

    @staticmethod
    def _get_metrics(problem: MetricsProblem) -> Dict[str, Metric]:
        """Returns metrics appropriate for the task, otherwise simply returns `metrics`.

        Args:
            problem: The problem type to return the appropriate metrics for.

        Returns:
            Dictionary of metrics to be computed.

        Raises:
            ValueError: If the problem type is not recognised.
        """
        if problem is MetricsProblem.REGRESSION:
            metrics = REGRESSION_METRICS
        elif problem is MetricsProblem.BINARY_CLASSIFICATION:
            metrics = BINARY_CLASSIFICATION_METRICS
        elif problem is MetricsProblem.MULTICLASS_CLASSIFICATION:
            metrics = MULTICLASS_CLASSIFICATION_METRICS
        elif problem is MetricsProblem.SEGMENTATION:
            metrics = SEGMENTATION_METRICS
        elif problem is MetricsProblem.MULTILABEL_CLASSIFICATION:
            metrics = MULTILABEL_CLASSIFICATION_METRICS
        else:
            raise ValueError("Problem type not recognised.")

        return metrics

    def _get_prediction_classes(
        self,
        test_target: np.ndarray,
        test_preds: np.ndarray,
        metric_to_optimise: str,
        threshold: Optional[float] = None,
    ) -> np.ndarray:
        """Returns test_preds as class labels.

        This is done after applying or optimising a threshold.
        """
        if (
            threshold is None
            and self.optimal_threshold is None
            and self.problem is not MetricsProblem.MULTICLASS_CLASSIFICATION
        ):
            self._optimise_threshold(test_target, test_preds, metric_to_optimise)

        return self._apply_threshold(test_preds, threshold)

    def _apply_threshold(
        self,
        preds: np.ndarray,
        threshold: Optional[float] = None,
    ) -> np.ndarray:
        """Applies a threshold to a list of predictions and returns the classes.

        For binary and multilabel problems, a threshold is computed and applied.

        For multiclass problems, it simply assumes that every prediction must
        have exactly 1 positive class and this is given to the class corresponding
        to the highest probability.
        """
        if threshold is None:
            threshold = self.optimal_threshold
        test_preds_cls: np.ndarray
        if self.problem is MetricsProblem.BINARY_CLASSIFICATION:
            test_preds_cls = np.where(preds < threshold, 0, 1)
        elif self.problem is MetricsProblem.MULTILABEL_CLASSIFICATION:
            test_preds_cls = np.asarray(
                [[0 if prob < threshold else 1 for prob in output] for output in preds]
            )
        else:
            test_preds_cls = np.array([np.argmax(output) for output in preds])
        return test_preds_cls

    def _optimise_threshold(
        self,
        test_target: np.ndarray,
        test_preds: np.ndarray,
        metric_to_optimise: str,
    ) -> None:
        """Optimise decision threshold by tuning metric_to_optimise.

        Finds the argmax of given metric by modifying the decision threshold and sets
        self.optimal_threshold

        Args:
            test_target: The targets array.
            test_preds: The predictions array.
            metric_to_optimise: The metric to optimise.

        Raises:
            ValueError: if metric_to_optimise does not exist in self.metrics
        """
        if metric_to_optimise not in self.metrics:
            raise ValueError("Chosen optimisation metric is not defined.")

        optimal_threshold = 0
        max_metric: float = 0

        # Dictionary of relevant metrics (i.e. those where a threshold is applied)
        # and the computed value for every threshold stored as a list
        threshold_metrics: Dict[str, List[float]] = {}
        for m in self.metrics:
            if (
                isinstance(self.metrics[m], ClassificationMetric)
                and not cast(ClassificationMetric, self.metrics[m]).probabilities
            ):
                threshold_metrics[m] = []
        # Loop through every threshold and apply it to our predictions
        for threshold in self.thresholds:
            test_preds_cls = self._apply_threshold(test_preds, threshold)

            # Loop through every metric in threshold_metrics and compute the value
            for metric_name in threshold_metrics:
                metric = self.metrics[metric_name]
                metric_val = metric.func(test_target, test_preds_cls)

                # Store the computed value in `threshold_metrics`
                threshold_metrics[metric_name].append(metric_val)

                # Update optimal threshold if the metric is the one we are optimising
                # and if the value is better than the previous best value
                if metric_name == metric_to_optimise and metric_val > max_metric:
                    max_metric = metric_val
                    optimal_threshold = threshold

        self.optimal_threshold = optimal_threshold
        self.threshold_metrics = threshold_metrics


def sum_fpr_curve(
    test_target: List[float],
    test_pred: List[float],
    test_var_to_sum: List[float],
    test_var_margin: Optional[List[float]] = None,
    granularity: int = 100,
) -> Tuple[List[float], List[float]]:
    """Returns the false positive rate to sums of a given variable.

    Returns this sum as a tuple of arrays (fpr, sums).

    This is calculated by moving a threshold that is used for deciding to
    accept/reject a given entry and then summing the associated variable. This
    can be used, for example, to calculate the total profit that would be
    achieved according to different false positive rates.

    Granularity is the number of regions to have that split up the threshold space.
    """
    thresholds = list(range(0, granularity + 1, 1))
    fprs = []
    test_vars = []

    for threshold in thresholds:
        test_preds_thresh = [
            0 if prob < threshold / granularity else 1 for prob in test_pred
        ]
        true_neg, false_pos, _, _ = confusion_matrix(
            test_target, test_preds_thresh
        ).ravel()
        fpr = 1 - (true_neg / (true_neg + false_pos))
        fprs.append(fpr)
        test_var = _sum_var(test_preds_thresh, test_var_to_sum, test_var_margin)
        test_vars.append(test_var)

    return fprs, test_vars


def _sum_var(
    test_preds_thresh: List[int],
    test_var_to_sum: List[float],
    test_var_margin: Optional[List[float]] = None,
) -> float:
    """Performs the required sum on the variable provided for every prediction."""
    cum_var: float = 0
    cum_margin: float = 0
    for index, val in enumerate(test_preds_thresh):
        if val == 1:
            cum_var += test_var_to_sum[index]
            if test_var_margin:
                cum_margin += test_var_margin[index]

    if test_var_margin:
        return cum_var / max(cum_margin, 1)

    return cum_var



"""Neural Network helper functions and classes."""

from abc import ABC, abstractmethod
from typing import Any


class _NeuralNet(ABC):
    """Base abstract class for all neural network models."""

    @abstractmethod
    def forward(self, x: Any) -> Any:
        """Run a forward pass of the model."""
        raise NotImplementedError


class _ConvNeuralNet(_NeuralNet, ABC):
    """Abstract class for all convolutional neural network models."""

    pass


class _FeedForwardNeuralNet(_NeuralNet, ABC):
    """Abstract class for all feed-forward neural network models."""

    pass



"""Tests for unary transformations."""
import copy
import logging
from typing import Optional, Type

from _pytest.logging import LogCaptureFixture
from marshmallow import ValidationError
import pytest

from bitfount.transformations.base_transformation import (
    MultiColumnOutputTransformation,
    Transformation,
)
from bitfount.transformations.unary_operations import (
    InclusionTransformation,
    OneHotEncodingTransformation,
    StringUnaryOperation,
)
from tests.bitfount.transformations.transformation_test_helpers import (
    gen_name_fails,
    gen_name_test,
    registration_test,
)
from tests.utils.helper import unit_test


def gen_name_test_unary(cls: Type[Transformation], registry_name: str) -> None:
    """Tests name generation for unary transformations.

    Tests that a unary transformation class is created with the correctly generated
    random name.
    """
    gen_name_test(cls, registry_name, arg="hello")


@unit_test
class TestStringUnaryOperation:
    """Tests for StringUnaryOperation abstract transformation."""

    def test__gen_name_fails(self) -> None:
        """Tests that name generation fails."""
        gen_name_fails(StringUnaryOperation)

    def test_schema_loads(self) -> None:
        """Test schema load works correctly."""
        s = StringUnaryOperation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg: hello
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg == "hello"

    def test_schema_load_fails_arg_nonstring(self) -> None:
        """Tests schema load fails if arg isn't a string."""
        s = StringUnaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg: 1
                """
            )

    def test_schema_load_fails_missing_arg(self) -> None:
        """Tests schema load fails if arg missing."""
        s = StringUnaryOperation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                """
            )


@unit_test
class TestOneHotEncodingTransformation:
    """Tests for the OneHotEncodingTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        # Test that name gen not attempted if no name and no reference provided.
        with pytest.raises(
            ValueError, match="No name provided and no reference found in arg."
        ):
            gen_name_test(
                OneHotEncodingTransformation,
                "onehotencode",
                arg="hello",
                raw_values=["blah"],
            )

        # Test that name gen attempted if reference provided.
        t = OneHotEncodingTransformation(arg="c:hello", raw_values=["blah"])
        assert t.name == "onehotencode_FAKE_HEX"

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(OneHotEncodingTransformation, "onehotencode")

    def test_is_multi_column_output(self) -> None:
        """Tests one-hot encoding correctly marked as multi-column output."""
        assert issubclass(OneHotEncodingTransformation, MultiColumnOutputTransformation)
        assert isinstance(
            OneHotEncodingTransformation(
                name="ohe_test", arg="test", raw_values=["blah"]
            ),
            MultiColumnOutputTransformation,
        )

    def test_init(self) -> None:
        """Tests that OHE is correctly initialized with various arg-types."""
        t_name: Optional[str]

        # With name
        t_name = "ohe_test"
        t_arg = "random_name"
        # With list `raw_values`
        ohe = OneHotEncodingTransformation(name=t_name, arg=t_arg, raw_values=[1, 2, 3])
        assert ohe.values == {i: f"{t_name}_{i}" for i in [1, 2, 3]}

        # With dict `raw_values`
        ohe = OneHotEncodingTransformation(
            name=t_name, arg=t_arg, raw_values={1: "hello", 2: "world"}
        )
        assert ohe.values == {
            k: f"{t_name}_{v}" for k, v in {1: "hello", 2: "world"}.items()
        }

        # With dict with Nones `raw_values`
        ohe = OneHotEncodingTransformation(
            name=t_name, arg=t_arg, raw_values={1: "hello", 2: None}
        )
        assert ohe.values == {
            k: f"{t_name}_{v}" for k, v in {1: "hello", 2: "2"}.items()
        }

        # With ref arg
        t_name = None
        t_ref = "ref_test"
        t_arg = f"c:{t_ref}"
        # With list `raw_values`
        ohe = OneHotEncodingTransformation(name=t_name, arg=t_arg, raw_values=[1, 2, 3])
        assert ohe.values == {i: f"{t_ref}_{i}" for i in [1, 2, 3]}

        # With dict `raw_values`
        ohe = OneHotEncodingTransformation(
            name=t_name, arg=t_arg, raw_values={1: "hello", 2: "world"}
        )
        assert ohe.values == {
            k: f"{t_ref}_{v}" for k, v in {1: "hello", 2: "world"}.items()
        }

        # With dict with Nones `raw_values`
        ohe = OneHotEncodingTransformation(
            name=t_name, arg=t_arg, raw_values={1: "hello", 2: None}
        )
        assert ohe.values == {
            k: f"{t_ref}_{v}" for k, v in {1: "hello", 2: "2"}.items()
        }

    def test_init_fails_duplicate_values(self) -> None:
        """Tests that initialization fails if duplicate values are provided."""
        # Test list `raw_values`
        with pytest.raises(
            ValueError, match="If `raw_values` is a list, elements must be unique."
        ):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values=[1, 1]
            )

        # Test dict `raw_values`
        with pytest.raises(ValueError, match="Column names generated must be unique:"):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values={1: "hello", 2: "hello"}
            )

        # Test dict with None `raw_values`
        with pytest.raises(ValueError, match="Column names generated must be unique:"):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values={1: "hello", "hello": None}
            )

    def test_init_fails_if_unknown_col_clash(self) -> None:
        """Tests that initialization fails if col clashes with unknown col."""
        # Test list `raw_values`
        with pytest.raises(
            ValueError,
            match="At least one column name clashes with the unknown value column",
        ):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values=["UNKNOWN"]
            )

        # Test dict `raw_values`
        with pytest.raises(
            ValueError,
            match="At least one column name clashes with the unknown value column",
        ):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values={1: "UNKNOWN"}
            )

        # Test dict with None `raw_values`
        with pytest.raises(
            ValueError,
            match="At least one column name clashes with the unknown value column",
        ):
            OneHotEncodingTransformation(
                name="ohe_test", arg="_unused", raw_values={"UNKNOWN": None}
            )

    def test_init_fails_if_no_name_and_non_ref_arg(self) -> None:
        """Tests the init fails if no name and non-reference arg."""
        with pytest.raises(
            ValueError, match="No name provided and no reference found in arg."
        ):
            OneHotEncodingTransformation(arg="not_a_ref", raw_values=["blah"])

    def test_init_fails_if_no_raw_values(self) -> None:
        """Tests the init fails if no raw_values are provided."""
        # Test with list raw_values
        with pytest.raises(
            ValueError, match="At least one value must be provided to one-hot encode."
        ):
            OneHotEncodingTransformation(name="name", arg="arg", raw_values=[])

        # Test with dict raw_values
        with pytest.raises(
            ValueError, match="At least one value must be provided to one-hot encode."
        ):
            OneHotEncodingTransformation(name="name", arg="arg", raw_values={})

    def test_columns_with_name_provided(self) -> None:
        """Tests the columns property when name provided."""
        t_name = "ohe_test"

        # Test when `raw_values` is a list
        ohe = OneHotEncodingTransformation(name=t_name, arg="", raw_values=[1, 2, 3])
        assert ohe.columns == [f"{t_name}_{i}" for i in [1, 2, 3]] + [
            f"{t_name}_UNKNOWN"
        ]

        # Test when `raw_values` is a dict
        ohe = OneHotEncodingTransformation(
            name=t_name, arg="", raw_values={1: "hello", 2: "world"}
        )
        assert ohe.columns == [f"{t_name}_{i}" for i in ["hello", "world"]] + [
            f"{t_name}_UNKNOWN"
        ]

        # Test when `raw_values` is a dict with Nones
        ohe = OneHotEncodingTransformation(
            name=t_name, arg="", raw_values={1: "hello", 2: None}
        )
        assert ohe.columns == [f"{t_name}_{i}" for i in ["hello", 2]] + [
            f"{t_name}_UNKNOWN"
        ]

    def test_columns_with_name_not_provided_but_arg_ref(self, fake_uuid: None) -> None:
        """Tests the columns property when no name provided but arg is a reference."""
        ref = "test"
        arg_ref = f"c:{ref}"

        # Test when `raw_values` is a list
        ohe = OneHotEncodingTransformation(arg=arg_ref, raw_values=[1, 2, 3])
        assert ohe.columns == [f"{ref}_{i}" for i in [1, 2, 3]] + [f"{ref}_UNKNOWN"]

        # Test when `raw_values` is a dict
        ohe = OneHotEncodingTransformation(
            arg=arg_ref, raw_values={1: "hello", 2: "world"}
        )
        assert ohe.columns == [f"{ref}_{i}" for i in ["hello", "world"]] + [
            f"{ref}_UNKNOWN"
        ]

        # Test when `raw_values` is a dict with Nones
        ohe = OneHotEncodingTransformation(
            arg=arg_ref, raw_values={1: "hello", 2: None}
        )
        assert ohe.columns == [f"{ref}_{i}" for i in ["hello", 2]] + [f"{ref}_UNKNOWN"]

    def test_prefix(self) -> None:
        """Tests prefix property."""
        # If name provided
        ohe = OneHotEncodingTransformation(
            name="ohe_test", arg="", raw_values=["_value"]
        )
        assert ohe.prefix == "ohe_test"

        # If no name but ref arg was provided
        ohe = OneHotEncodingTransformation(arg="c:test", raw_values=["_value"])
        assert ohe.prefix == "test"

        # If no name or ref arg was provided
        with pytest.raises(
            ValueError, match="No name provided and no reference found in arg."
        ):
            OneHotEncodingTransformation(arg="not_a_ref", raw_values=["_value"])

    def test_unknown_col(self) -> None:
        """Tests unknown_col property."""
        ohe = OneHotEncodingTransformation(
            name="ohe_test", arg="", raw_values=["_value"]
        )
        assert ohe.unknown_col == f"{ohe.prefix}_UNKNOWN"

    def test_unknown_col_custom_suffix(self) -> None:
        """Tests unknown_col property with custom suffix."""
        ohe = OneHotEncodingTransformation(
            name="ohe_test", arg="", raw_values=["_value"], unknown_suffix="CUSTOM"
        )
        assert ohe.unknown_col == f"{ohe.prefix}_CUSTOM"

    def test_trying_to_init_values_twice(self, caplog: LogCaptureFixture) -> None:
        """Tests that values cannot be set twice."""
        ohe = OneHotEncodingTransformation(
            name="ohe_test", arg="_unused", raw_values=[1, 2, 3]
        )
        orig_values = copy.deepcopy(ohe.values)

        # Try to init again
        with caplog.at_level(logging.WARNING):
            ohe._produce_full_col_map()

        log_record = caplog.records[0]
        assert log_record.levelname == "WARNING"
        assert (
            log_record.message
            == "_produce_full_col_map should not be called more than once"
        )
        assert orig_values == ohe.values

    # The tests below this point are all related to the schema validation. It
    # duplicates many of the tests in the transformation hierarchy but is needed
    # here because OneHotEncodingTransformation has a custom schema.
    def test_schema_loads_list_values(self) -> None:
        """Test schema load works correctly with list values."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg: hello
            values: [ 1, 2, 3 ]
            unknown_suffix: UNKNOWN_TEST
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg == "hello"
        assert t.values == {i: f"test_{i}" for i in [1, 2, 3]}
        assert t.unknown_suffix == "UNKNOWN_TEST"

    def test_schema_loads_dict_values(self) -> None:
        """Test schema load works correctly with dict values."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg: hello
            values: { 1: "hello", 2: "world" }
            unknown_suffix: UNKNOWN_TEST
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg == "hello"
        assert t.values == {k: f"test_{v}" for k, v in {1: "hello", 2: "world"}.items()}
        assert t.unknown_suffix == "UNKNOWN_TEST"

    def test_schema_loads_dict_null_values(self) -> None:
        """Test schema load works correctly with null dict values."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            name: test
            output: True
            arg: hello
            values:
              1: hello
              2:
            unknown_suffix: UNKNOWN_TEST
            """
        )
        assert t.name == "test"
        assert t.output is True
        assert t.arg == "hello"
        assert t.values == {k: f"test_{v}" for k, v in {1: "hello", 2: "2"}.items()}
        assert t.unknown_suffix == "UNKNOWN_TEST"

    def test_schema_load_succeeds_without_name(self, fake_uuid: None) -> None:
        """Tests schema loads successfully without name field."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            output: True
            arg: c:hello
            values: [ 1, 2, 3 ]
            """
        )
        assert t.name == "onehotencode_FAKE_HEX"
        assert t.output is True

    def test_schema_load_fails_with_wrong_type_name(self) -> None:
        """Tests schema load fails if non-str name field."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: 2
                arg: hello
                values: [ 1, 2, 3 ]
                """
            )

    def test_schema_load_succeeds_without_output(self) -> None:
        """Tests schema loads successfully without output field."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            name: test
            arg: hello
            values: [ 1, 2, 3 ]
            """
        )
        assert t.name == "test"
        assert t.output is False

    def test_schema_load_fails_with_wrong_type_output(self) -> None:
        """Tests schema load fails if non-bool output field."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: hello
                arg: hello
                values: [ 1, 2, 3 ]
                """
            )

    def test_schema_load_fails_missing_arg(self) -> None:
        """Tests schema load fails if arg missing."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                values: [ 1, 2, 3 ]
                """
            )

    def test_schema_load_fails_arg_nonstring(self) -> None:
        """Tests schema load fails if arg isn't a string."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                arg: 1
                values: [ 1, 2, 3 ]
                """
            )

    def test_schema_load_fails_missing_values(self) -> None:
        """Tests schema load fails if values is missing."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                arg: hello
                """
            )

    def test_schema_load_fails_values_wrong_type(self) -> None:
        """Tests schema load fails if values is an incompatible type."""
        s = OneHotEncodingTransformation.schema()

        # Values is non-list, non-dict
        with pytest.raises(ValidationError):
            s.loads(
                """
                arg: hello
                values: 1
                """
            )

        # Values is dict but with non-str values
        with pytest.raises(ValidationError):
            s.loads(
                """
                arg: hello
                values: { 1: 1 }
                """
            )

        # Values is null
        with pytest.raises(ValidationError):
            s.loads(
                """
                arg: hello
                values: null
                """
            )

    def test_schema_load_succeeds_without_unknown_suffix(self) -> None:
        """Tests schema loads successfully without unknown_suffix field."""
        s = OneHotEncodingTransformation.schema()
        t = s.loads(
            """
            name: test
            arg: hello
            values: [ 1, 2, 3 ]
            """
        )
        assert t.unknown_suffix == "UNKNOWN"

    def test_schema_load_fails_unknown_suffix_non_string(self) -> None:
        """Tests schema load fails if unknown_suffix isn't a string."""
        s = OneHotEncodingTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                arg: hello
                values: [ 1, 2, 3 ]
                unknown_suffix: false
                """
            )


@unit_test
class TestInclusionTransformation:
    """Tests for the InclusionTransformation class."""

    def test__gen_name(self, fake_uuid: None) -> None:
        """Tests transformation name generation."""
        gen_name_test(InclusionTransformation, "in", arg="hello", in_str="hell")

    def test_registration(self) -> None:
        """Tests transformation registration."""
        registration_test(InclusionTransformation, "in")

    def test_schema_load_fails_arg2_nonstring(self) -> None:
        """Tests schema load fails if arg2 is not a string."""
        s = InclusionTransformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: True
                arg1: hello
                arg2: 2
                """
            )



"""Model implementations using PyTorch."""
from __future__ import annotations

import logging
from typing import Any, ClassVar, List, Optional, Tuple, Union, cast

from marshmallow import fields
import pandas as pd
from pytorch_tabnet.tab_model import TabNetClassifier as TabNetClassifier_
import torch
from torch import nn as nn
from torch.nn import Module

from bitfount.backends.pytorch.data.dataloaders import (
    PyTorchBitfountDataLoader,
    _BasePyTorchBitfountDataLoader,
)
from bitfount.backends.pytorch.models.base_models import (
    _OPTIMIZERS,
    _SCHEDULERS,
    BasePyTorchModel,
    BaseTabNetModel,
    PyTorchClassifierMixIn,
    _calculate_embedding_sizes,
)
from bitfount.backends.pytorch.models.nn import (
    _get_torchvision_classification_model,
    _PyTorchConvNeuralNet,
    _PyTorchFeedForwardNeuralNet,
    _PyTorchLogisticRegression,
)
from bitfount.backends.pytorch.types import (
    ImgAndTabDataSplit,
    ImgDataReturnType,
    ImgFwdTypes,
    ImgXorTabDataSplit,
    TabDataReturnType,
)
from bitfount.data.datastructure import DataStructure
from bitfount.data.types import SemanticType
from bitfount.models.base_models import (
    CNNModelStructure,
    FeedForwardModelStructure,
    NeuralNetworkModelStructure,
    NeuralNetworkPredefinedModel,
    Scheduler,
)
from bitfount.types import T_FIELDS_DICT
from bitfount.utils import DEFAULT_SEED, delegates

logger = logging.getLogger(__name__)


@delegates()
class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
    """A Pytorch model designed specifically for image classification problems.

    The model can handle binary, multiclass and multilabel classification problems.

    :::info

    Currently images are scaled to 224 x 224.

    :::

    Raises:
        ValueError: If the model structure is not an instance of `CNNModelStructure` or
            `NeuralNetworkPredefinedModel`.
        ValueError: If the loss function is not `nn.BCEWithLogitsLoss` or
            `nn.CrossEntropyLoss`.
    """

    train_dl: _BasePyTorchBitfountDataLoader
    datastructure: DataStructure
    input_dim: int = 224  # default input images are 224 x 224

    def __init__(self, **kwargs: Any):
        model_structure = kwargs.pop("model_structure", CNNModelStructure())
        if not isinstance(
            model_structure, (CNNModelStructure, NeuralNetworkPredefinedModel)
        ):
            raise ValueError("Model structure does not match model")

        super().__init__(model_structure=model_structure, **kwargs)
        if self.loss_func is None:
            if self.multilabel:
                self.loss_func = nn.BCEWithLogitsLoss
            else:
                self.loss_func = nn.CrossEntropyLoss
        elif self.loss_func not in [nn.BCEWithLogitsLoss, nn.CrossEntropyLoss]:
            raise ValueError("This loss function is not currently supported")

    def forward(self, x: ImgFwdTypes) -> Any:  # type: ignore[override] # Reason: see below # noqa: B950
        """Performs a forward pass of the model."""
        # override as the forward function is incompatible with pl.LightningModule
        if self.datastructure.number_of_images > 1:
            aux = []
            for i in range(len(x)):
                aux.append(self._model(x[i]))  # type: ignore[misc] # reason: model should be initialised already # noqa: B950
            return torch.cat([item[0] for item in aux], 1)
        else:
            return self._model(x)  # type: ignore[misc] # reason: model should be initialised already # noqa: B950

    def _split_dataloader_output(
        self,
        data: Union[
            ImgAndTabDataSplit,
            ImgXorTabDataSplit,
        ],
    ) -> Union[ImgDataReturnType, TabDataReturnType]:
        """Splits dataloader output into image tensor, weights and category."""
        images, sup = cast(Tuple[torch.Tensor, torch.Tensor], data)
        weights = sup[:, 0].float()
        category: Optional[torch.Tensor]
        if sup.shape[1] > 2:
            category = sup[:, -1:].long()
        else:
            category = None

        return images, weights, category

    def _get_convolution_final_output_dimension(self) -> int:
        """Calculates the output size of the final convolutional layer.

        This will become the input size of the first feedforward layer
        """
        input_dim: float = self.input_dim
        if isinstance(self.model_structure, CNNModelStructure):
            # self.model_structure.layers is set in post_init, so we can just cast
            for _ in cast(List[int], self.model_structure.layers):
                output_dim = self._get_convolution_output_dimension(
                    input_dim,
                    self.model_structure.kernel_size,
                    self.model_structure.padding,
                    self.model_structure.stride,
                )
                output_dim = output_dim / 2  # due to pooling
                input_dim = output_dim

            return int(
                (output_dim**2) * cast(List[int], self.model_structure.layers)[-1]
            )
        else:
            raise TypeError("This method only works with the cnn model structure.")

    @staticmethod
    def _get_convolution_output_dimension(
        input_size: Union[int, float], kernel_size: int, padding: int, stride: int
    ) -> float:
        """Gets convolution output dimension."""
        return ((input_size - kernel_size + (2 * padding)) / stride) + 1

    def _create_model(self) -> nn.Module:
        """Creates the model to use.

        If `self.model_structure` is a `NeuralNetworkPredefinedModel`, then calls
        `get_torchvision_classification_model` to adapt model head for the task
        before returning the model

        Otherwise, creates model as defined by `CNNModelStructure`
        """
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )

        self.set_number_of_classes(table_schema)

        if isinstance(self.model_structure, NeuralNetworkPredefinedModel):
            kwargs = self.model_structure.kwargs or {}
            model = _get_torchvision_classification_model(
                self.model_structure.name,
                self.model_structure.pretrained,
                self.n_classes,
                **kwargs,
            )
        elif isinstance(self.model_structure, CNNModelStructure):
            # self.model_structure.layers is set in post_init, so we can just cast
            layer_sizes = self._get_layer_sizes(
                cast(List[int], self.model_structure.layers), 3
            )

            head_sizes = [
                (self.model_structure.ff_layers[-1], self.n_classes)
                for _ in range(self.model_structure.num_heads)
            ]
            ff_layer_sizes = self._get_layer_sizes(
                self.model_structure.ff_layers,
                self._get_convolution_final_output_dimension(),
            )
            logger.debug(f"Creating model with {self.model_structure.num_heads} heads")
            # self.model_structure.dropout_probs is set in post_init,
            # so we can just cast below
            model = _PyTorchConvNeuralNet(
                layer_sizes,
                cast(List[float], self.model_structure.dropout_probs),
                self.model_structure.mish_activation_function,
                head_sizes,
                ff_layer_sizes,
                self.model_structure.ff_dropout_probs,
                self.model_structure.kernel_size,
                self.model_structure.padding,
                self.model_structure.stride,
                self.model_structure.pooling_function,
            )
        return model


@delegates()
class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
    """A Pytorch model designed specifically for tabular classification problems.

    The model can handle binary, multiclass and multilabel classification problems.

    Raises:
        ValueError: If the model structure is not an instance of
            `FeedForwardModelStructure`.
        ValueError: If the loss function is not `nn.BCEWithLogitsLoss` or
            `nn.CrossEntropyLoss`.
    """

    train_dl: _BasePyTorchBitfountDataLoader

    def __init__(self, **kwargs: Any):
        model_structure = kwargs.pop("model_structure", FeedForwardModelStructure())
        if (
            isinstance(model_structure, NeuralNetworkPredefinedModel)
            and model_structure.name == "TabNet"
        ):
            raise ValueError("Please create a TabNetClassifier directly.")
        elif not isinstance(model_structure, FeedForwardModelStructure):
            raise ValueError("Please provide a FeedForwardModelStructure")
        super().__init__(
            model_structure=cast(NeuralNetworkModelStructure, model_structure),
            **kwargs,
        )

        if self.loss_func is None:
            if hasattr(self, "multilabel") and (self.multilabel is not False):
                self.loss_func = nn.BCEWithLogitsLoss
            else:
                self.loss_func = nn.CrossEntropyLoss
        elif self.loss_func not in [nn.BCEWithLogitsLoss, nn.CrossEntropyLoss]:
            raise ValueError("This loss function is not currently supported")

    def _create_model(self) -> _PyTorchFeedForwardNeuralNet:
        """Creates model to use.

        Takes number of continuous features and number of heads. Creates and
        returns model.
        """
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )
        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
            table_schema
        )

        self.set_number_of_classes(table_schema)
        num_continuous = len(
            [
                col
                for col in table_schema.get_feature_names(
                    SemanticType.CONTINUOUS,
                )
                if col not in ignore_cols_for_training
            ]
        )
        embedding_sizes = _calculate_embedding_sizes(
            table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
        )
        self.model_structure = cast(FeedForwardModelStructure, self.model_structure)
        num_heads = self.model_structure.num_heads
        num_categorical = sum(size for _, size in embedding_sizes)
        # self.model_structure.layers is set in post_init, so we can just cast
        layer_sizes = self._get_layer_sizes(
            cast(
                List[int],
                self.model_structure.layers,
            ),
            num_continuous + num_categorical,
        )
        head_sizes = [
            (
                cast(
                    List[int],
                    self.model_structure.layers,
                )[-1],
                self.n_classes,
            )
            for _ in range(num_heads)
        ]
        # self.model_structure.dropout_probs is set in post_init,
        # so we can just cast below
        logger.debug(f"Creating model with {num_heads} heads")
        model = _PyTorchFeedForwardNeuralNet(
            embedding_sizes,
            self.model_structure.embedding_dropout,
            num_continuous,
            layer_sizes,
            cast(List[float], self.model_structure.dropout_probs),
            self.model_structure.mish_activation_function,
            head_sizes,
        )
        return model

    def _split_dataloader_output(
        self,
        data: Union[
            ImgAndTabDataSplit,
            ImgXorTabDataSplit,
        ],
    ) -> Union[ImgDataReturnType, TabDataReturnType]:
        """Splits dataloader output.

        Splits it into pieces for categorical, continuous, weights and categories.

        NB: `ignore_classes` is never returned
        """
        tab, sup = cast(ImgXorTabDataSplit, data)

        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )
        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
            table_schema
        )

        n_cat = len(
            _calculate_embedding_sizes(
                table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
            )
        )
        n_cont = len(
            [
                col
                for col in table_schema.get_feature_names(
                    SemanticType.CONTINUOUS,
                )
                if col not in ignore_cols_for_training
            ]
        )
        # Get items according to the order they are in the tensor
        cat_pos = n_cat
        cont_pos = cat_pos + n_cont
        x_1 = tab[:, :cat_pos].long()  # categorical features
        x_2 = tab[:, cat_pos:cont_pos].float()  # continuous features
        weights = sup[:, 0].float()
        # If category is present, return it, otherwise return None
        category: Optional[torch.Tensor]
        if sup.shape[1] > 2:
            category = sup[:, -1:].long()
        else:
            category = None
        return (x_1.t(), x_2), weights, category


@delegates()
class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
    """A Logistic Regression classifier implemented in PyTorch.

    Utilises softmax regression to allow extension to more than 2 classes.
    The input and output dimensions are calculated from the data automatically.
    A softmax regressor is used to ensure this can work on multi-class
        problems.

    Args:
        bias: Whether the underlying linear layer should learn an additive
            bias. Default True.
        l1_regularization_weight: The weight of L1 regularization to apply, if any.
        l2_regularization_weight: The weight of L2 regularization to apply, if any.
        embed_categorical: Whether to use categorical embedding layers to handle
            categorical variables, or to treat them as inherently label-encoded.
            Default True.
        embed_categorical_dropout: The dropout probability to apply to categorical
            variables if using embed_categorical.
        **kwargs: Other keyword arguments for the model.
    """

    train_dl: PyTorchBitfountDataLoader
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "bias": fields.Boolean(),
        "l1_regularization_weight": fields.Float(allow_none=True),
        "l2_regularization_weight": fields.Float(allow_none=True),
        "embed_categorical": fields.Boolean(),
        "embed_categorical_dropout": fields.Float(),
    }

    def __init__(
        self,
        bias: bool = True,
        l1_regularization_weight: Optional[float] = None,
        l2_regularization_weight: Optional[float] = None,
        embed_categorical: bool = True,
        embed_categorical_dropout: float = 0.04,
        **kwargs: Any,
    ) -> None:
        """Create a logistic regression classifier using PyTorch."""
        # Override model_structure; model_structure is required by the
        # NeuralNetworkMixin, but is not used in this class, so we override it
        # with an empty one and log out that this has happened
        # Have to use empty lists as otherwise they get overwritten with defaults
        empty_model_structure = FeedForwardModelStructure(
            layers=[], dropout_probs=[], mish_activation_function=False
        )
        try:
            if kwargs["model_structure"] != empty_model_structure:
                logger.warning(
                    f"Specified model structure is not compatible with "
                    f"{self.__class__.__name__}; will be overridden."
                )
                kwargs["model_structure"] = empty_model_structure
        except KeyError:
            kwargs["model_structure"] = empty_model_structure

        super().__init__(**kwargs)

        # Multilabel LogReg is not supported
        if self.multilabel:
            raise ValueError(
                f"{self.__class__.__name__} does not support multilabel "
                f"classification problems."
            )
        # Multihead LogReg is not supported
        try:
            model_structure = cast(NeuralNetworkModelStructure, self.model_structure)
            if model_structure.num_heads != 1:
                logger.warning("Multihead LogReg is not supported, setting to 1.")
                model_structure.num_heads = 1
        except AttributeError:
            pass

        # Set loss function. Only CrossEntropyLoss is currently supported.
        if self.loss_func is None:
            self.loss_func = nn.CrossEntropyLoss
        elif self.loss_func is not nn.CrossEntropyLoss:
            raise ValueError("This loss function is not currently supported")

        self.bias = bias
        self.l1_regularization_weight = l1_regularization_weight
        self.l2_regularization_weight = l2_regularization_weight
        self.embed_categorical = embed_categorical
        self.embed_categorical_dropout = embed_categorical_dropout

    def _create_model(self) -> nn.Module:
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )
        # Set number of output classes from datastructure
        self.set_number_of_classes(table_schema)

        # Get details on continuous and categorical features
        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
            table_schema
        )

        if self.embed_categorical:
            # If categorical embedding requested, calculate embedding sizes
            logger.info("Generating Categorical Embeddings for categorical features.")
            embedding_sizes = _calculate_embedding_sizes(
                table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
            )
            num_continuous = table_schema.get_num_continuous(
                ignore_cols=ignore_cols_for_training
            )
            model = _PyTorchLogisticRegression(
                num_classes=self.n_classes,
                num_continuous=num_continuous,
                embedding_sizes=embedding_sizes,
                embedding_dropout_frac=self.embed_categorical_dropout,
                bias=self.bias,
            )
        else:
            # Treat categorical values as label encoded
            logger.info(
                "No categorical embedding requested;"
                " categorical data should be label-encoded."
            )
            num_categorical = table_schema.get_num_categorical(
                ignore_cols=ignore_cols_for_training
            )
            num_continuous = table_schema.get_num_continuous(
                ignore_cols=ignore_cols_for_training
            )
            input_dim = num_continuous + num_categorical
            model = _PyTorchLogisticRegression(
                num_classes=self.n_classes,
                input_dim=input_dim,
                bias=self.bias,
            )

        return model

    def _split_dataloader_output(
        self,
        data: Union[
            ImgAndTabDataSplit,
            ImgXorTabDataSplit,
        ],
    ) -> Union[ImgDataReturnType, TabDataReturnType]:
        """Splits dataloader output.

        Splits it into pieces for categorical, continuous, weights and
        (optionally) categories.

        Depending on whether embed_categorical is True or not will affect the shape
        of the returned categorical data tensor, either to ensure it can pass through
        the embedding layers correctly or to treat it as inherently label-encoded.

        NB: `ignore_classes` is never returned
        """
        # Split into the tabular data and supplementary data
        tab, sup = cast(ImgXorTabDataSplit, data)

        # Gather schema-related information
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )
        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
            table_schema
        )

        # Calculate the number of continuous and categorical columns
        n_categorical = table_schema.get_num_categorical(
            ignore_cols=ignore_cols_for_training
        )
        n_continuous = table_schema.get_num_continuous(
            ignore_cols=ignore_cols_for_training
        )

        # Split tensor into categorical and continuous items
        end_categorical_idx = n_categorical
        x_categorical = tab[:, :end_categorical_idx].long()  # categorical features
        end_continuous_idx = end_categorical_idx + n_continuous
        x_continuous = tab[
            :, end_categorical_idx:end_continuous_idx
        ].float()  # continuous features

        # Weights will be first entry in the supplementary tensor
        weights = sup[:, 0].float()

        # If category is present (the remaining entries in the supplementary tensor),
        # return it, otherwise return None
        category: Optional[torch.Tensor]
        if sup.shape[1] > 2:
            category = sup[:, -1:].long()
        else:
            category = None

        if self.embed_categorical:
            # Return (transposed categorical tensor, continuous tensor), weights
            # tensor and category tensor (or None).
            # Categorical must be transposed to allow it to pass through the
            # embedding layers correctly.
            return (x_categorical.t(), x_continuous), weights, category
        else:
            # Otherwise, return categorical as is, expecting it to be label-encoded.
            # Return (categorical tensor, continuous tensor), weights tensor and
            # category tensor (or None).
            return (x_categorical, x_continuous), weights, category

    def _get_loss(
        self,
        output: torch.Tensor,
        target: torch.Tensor,
        loss_modifiers: Tuple[torch.Tensor, ...],
    ) -> torch.Tensor:
        """Computes the appropriate (weighted) loss.

        Applies L1 and L2 regularization if requested.

        Args:
            output: The model output.
            target: The expected output.
            loss_modifiers: A tuple of tensors representing additional loss modifiers:
                - 0: weightings for each sample in this batch
                - 1: categories

        Returns:
            A scalar tensor representing the loss.
        """
        loss = super()._get_loss(output, target, loss_modifiers)

        sample_weight: torch.Tensor = loss_modifiers[0]
        weight_sum = sample_weight.squeeze().sum()

        # Apply L1 regularisation if requested
        if (
            self.l1_regularization_weight is not None
            and self.l1_regularization_weight > 0
        ):
            model: Module = cast(Module, self._model)
            linear_layer: Module = cast(Module, model.linear)
            weight_tensor: torch.Tensor = cast(torch.Tensor, linear_layer.weight)

            l1_reg_term = weight_tensor.abs().sum()
            # As loss is already a (weighted) average we need to similarly divide the
            # regularisation term by the total weight.
            l1_reg_term /= weight_sum
            loss += self.l1_regularization_weight * l1_reg_term

        # Apply L2 regularisation if requested
        if (
            self.l2_regularization_weight is not None
            and self.l2_regularization_weight > 0
        ):
            model = cast(Module, self._model)
            linear_layer = cast(Module, model.linear)
            weight_tensor = cast(torch.Tensor, linear_layer.weight)

            l2_reg_term = weight_tensor.pow(2).sum()
            # As loss is already a (weighted) average we need to similarly divide the
            # regularisation term by the total weight.
            l2_reg_term /= weight_sum
            loss += self.l2_regularization_weight * l2_reg_term

        return loss


@delegates()
class TabNetClassifier(PyTorchClassifierMixIn, BaseTabNetModel):
    """TabNet Classifier for binary and multiclass tabular classification problems.

    See base class for more information.
    """

    train_dl: PyTorchBitfountDataLoader

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _create_model(self) -> TabNetClassifier_:
        """Create model for Binary or Multiclass classification."""
        table_schema = self.datastructure.get_table_schema(
            schema=self.schema, pod_identifier=self.pod_identifier
        )
        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
            table_schema
        )

        self.set_number_of_classes(table_schema)

        # Only consider the tabular part
        X: pd.DataFrame
        if isinstance(self.train_dl.get_x_dataframe(), tuple):
            X, _ = self.train_dl.get_x_dataframe()
        elif isinstance(self.train_dl.get_x_dataframe(), pd.DataFrame):
            X = cast(pd.DataFrame, self.train_dl.get_x_dataframe())
        cat_idxs = [
            i
            for i, f in enumerate(X.columns)
            if f
            in table_schema.get_feature_names(
                SemanticType.CATEGORICAL,
            )
            if f not in ignore_cols_for_training
        ]

        if self.embedding_sizes is None:

            embedding_sizes = _calculate_embedding_sizes(
                table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
            )
            self.embedding_sizes = cast(List[int], [i[1] for i in embedding_sizes])

        return TabNetClassifier_(
            cat_idxs=cat_idxs,
            cat_dims=table_schema.get_categorical_feature_sizes(
                ignore_cols=ignore_cols_for_training
            ),
            cat_emb_dim=self.embedding_sizes,
            optimizer_fn=_OPTIMIZERS[self.optimizer.name],
            optimizer_params=self.optimizer.params,
            scheduler_params=cast(Scheduler, self.scheduler).params,
            scheduler_fn=_SCHEDULERS[cast(Scheduler, self.scheduler).name],
            mask_type=self.mask_type,
            seed=self.seed or DEFAULT_SEED,
        )



"""Results Only protocol."""
from __future__ import annotations

import time
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    List,
    Optional,
    Protocol,
    Tuple,
    cast,
    runtime_checkable,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from marshmallow import fields
import pandas as pd

from bitfount.data.datasources.base_source import BaseSource
import bitfount.federated.algorithms.base as algorithms
from bitfount.federated.exceptions import PSINoDataSourceError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.pod_vitals import _PodVitals
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.protocols.base import (
    _BaseCompatibleAlgoFactory,
    _BaseModellerProtocol,
    _BaseProtocolFactory,
    _BaseWorkerProtocol,
)
from bitfount.federated.transport.modeller_transport import (
    _get_psi_datasets_from_workers,
    _get_public_key,
    _ModellerMailbox,
    _send_psi_dataset_modeller,
)
from bitfount.federated.transport.worker_transport import (
    _get_psi_dataset,
    _send_psi_dataset_worker,
    _send_public_key,
    _WorkerMailbox,
)
from bitfount.federated.types import SerializedProtocol
from bitfount.schemas.utils import bf_dump
from bitfount.types import T_FIELDS_DICT, T_NESTED_FIELDS
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub

logger = _get_federated_logger(__name__)


@runtime_checkable
class _PSICompatibleModeller(Protocol):
    """Defines modeller-side algorithm compatibility."""

    def initialise(
        self,
        datasource: BaseSource,
        public_key: RSAPublicKey,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        **kwargs: Any,
    ) -> List[int]:
        """Initialises the modeller-side algorithm."""
        ...

    def run(self, pod_set: List[int], modeller_set: List[int]) -> pd.DataFrame:
        """Runs the modeller-side algorithm."""
        ...


@runtime_checkable
class _PSICompatibleWorker(Protocol):
    public_key: RSAPublicKey

    def initialise(
        self,
        datasource: BaseSource,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        **kwargs: Any,
    ) -> List[int]:
        """Initialises the worker-side algorithm."""
        ...

    def run(self, modeller_set: List[int]) -> List[int]:
        """Runs the worker-side algorithm."""
        ...


class _ModellerSide(_BaseModellerProtocol):
    """Modeller side of the PrivateSetIntersection protocol."""

    algorithm: _PSICompatibleModeller
    datasource: BaseSource

    def __init__(
        self,
        *,
        algorithm: _PSICompatibleModeller,
        datasource: Optional[BaseSource] = None,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        mailbox: _ModellerMailbox,
        **kwargs: Any,
    ):
        if datasource is None:
            # Datasource is optional to match the parent class signature.
            raise PSINoDataSourceError(
                "You are trying to run a PSI task with no datasource. "
                "Please provide a datasource and try again."
            )
        self.datasource = datasource
        self.columns_to_intersect = columns_to_intersect
        self.table = table
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)

    async def _send_psi_data_modeller(self, dataset: List[int]) -> None:
        """Sends psi dataset to workers."""
        logger.debug("Sending PSI dataset to the worker.")
        # Change integers to strings so msgpack can handle them.
        str_dataset = [str(item) for item in dataset]
        await _send_psi_dataset_modeller(str_dataset, self.mailbox)

    async def _receive_public_key(self) -> RSAPublicKey:
        """Receives public key from worker."""
        logger.debug("Receiving PSI parameters")
        serialized_keys: List[bytes] = await _get_public_key(self.mailbox)
        # We only support psi with one pod, so get the first key only.
        serialized_key = serialized_keys[0]
        # Cast to RSAPublicKEy, as it is the expected key type.
        return cast(RSAPublicKey, load_pem_public_key(serialized_key))

    async def _receive_psi_datasets_from_workers(self) -> Tuple[List[int], List[int]]:
        """Receives psi datasets from worker."""
        logger.debug("Receiving PSI datasets")
        psi_datasets: List[
            Tuple[List[str], List[str]]
        ] = await _get_psi_datasets_from_workers(self.mailbox)
        # We only support psi with one pod, so get the first item of the list.
        psi_datasets_str = psi_datasets[0]
        # Change dataset to int.
        # The first list of the tuple is the pod set, and the second is the modeller set
        psi_datasets_int = (
            [int(item) for item in psi_datasets_str[0]],
            [int(item) for item in psi_datasets_str[1]],
        )
        return psi_datasets_int

    async def run(
        self,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Runs Modeller side of the protocol."""
        public_key = await self._receive_public_key()
        modeller_input = self.algorithm.initialise(
            public_key=public_key,
            datasource=self.datasource,
            columns_to_intersect=self.columns_to_intersect,
            table=self.table,
        )
        await self._send_psi_data_modeller(modeller_input)
        pod_set, modeller_set = await self._receive_psi_datasets_from_workers()
        modeller_results = self.algorithm.run(pod_set, modeller_set)
        await self.mailbox.send_task_complete_message()
        return modeller_results


class _WorkerSide(_BaseWorkerProtocol):
    """Worker side of the PrivateSetIntersection protocol."""

    algorithm: _PSICompatibleWorker

    def __init__(
        self,
        *,
        algorithm: _PSICompatibleWorker,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        mailbox: _WorkerMailbox,
        **kwargs: Any,
    ):
        self.columns_to_intersect = columns_to_intersect
        self.table = table
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)

    async def _receive_modeller_psi_dataset(self) -> List[int]:
        """Receives psi dataset from modeller."""
        logger.debug("Receiving PSI dataset from modeller.")
        str_dataset = await _get_psi_dataset(self.mailbox)
        return [int(item) for item in str_dataset]

    async def _send_public_key_to_modeller(self, public_key: RSAPublicKey) -> None:
        """Sends the public key for psi to modeller."""
        await _send_public_key(public_key, self.mailbox)

    async def _send_psi_data_to_modeller(
        self, dataset: Tuple[List[int], List[int]]
    ) -> None:
        """Sends the psi datasets back to the modeller."""
        # Change integers to strings so msgpack can handle them.
        str_tuple = (
            [str(item) for item in dataset[0]],
            [str(item) for item in dataset[1]],
        )
        await _send_psi_dataset_worker(str_tuple, self.mailbox)

    async def run(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_vitals: Optional[_PodVitals] = None,
        pod_identifier: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Runs Worker side of the protocol."""
        await self._send_public_key_to_modeller(self.algorithm.public_key)
        pod_set = self.algorithm.initialise(
            datasource=datasource,
            columns_to_intersect=self.columns_to_intersect,
            table=self.table,
        )
        modeller_set = await self._receive_modeller_psi_dataset()
        if pod_vitals:
            pod_vitals.last_task_execution_time = time.time()
        hashed_modeller_set = self.algorithm.run(modeller_set)
        await self._send_psi_data_to_modeller((pod_set, hashed_modeller_set))
        await self.mailbox.get_task_complete_update()


@runtime_checkable
class _PSICompatibleAlgoFactory_(_BaseCompatibleAlgoFactory, Protocol):
    """Defines algo factory compatibility."""

    def modeller(self, **kwargs: Any) -> _PSICompatibleModeller:
        """Creates a modeller-side algorithm."""
        ...

    def worker(self, **kwargs: Any) -> _PSICompatibleWorker:
        """Creates a worker-side algorithm."""
        ...


@delegates()
class PrivateSetIntersection(_BaseProtocolFactory):
    """Performs a private set intersection with the provided algorithm.

    For more details, look at the algorithm description to
    understand how the intersection is computed.

    Args:
        algorithm: The algorithm to run.
        datasource: The modeller's datasource.
        datasource_columns: The modeller's columns from their datasource
            on which the private set intersection will be computed as a
            list of strings. Defaults to None.
        datasource_table: The modeller's table from their datasource,
            if the datasource is multitable, on which the private set
            intersection will be computed as a string. Defaults to None.
        pod_columns: The pod's columns from their datasource
            on which the private set intersection will be computed as a
            list of strings. Defaults to None.
        pod_table: The pod's table from their datasource,
            if the datasource is multitable, on which the private set
            intersection will be computed as a string. Defaults to None.

    Attributes:
        name: The name of the protocol.
        algorithm: The algorithm to run. This must be compatible with the
            `PrivateSetIntersection` protocol.
        pod_columns: The pod's columns from their datasource
            on which the private set intersection will be computed as a
            list of strings. Defaults to None.
        pod_table: The pod's table from their datasource,
            if the datasource is multitable, on which the private set
            intersection will be computed as a string. Defaults to None.

    Raises:
        TypeError: If the `algorithm` is not compatible with the protocol.
        PSINoDataSourceError: If you are trying to run the protocol without
            a datasource.
    """

    algorithm: _PSICompatibleAlgoFactory_
    # We only need to serialize the pod columns & table
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "pod_columns": fields.List(fields.String(default=None), allow_none=True),
        "pod_table": fields.String(allow_none=True),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"algorithm": algorithms.registry}

    def __init__(
        self,
        *,
        algorithm: _PSICompatibleAlgoFactory_,
        datasource: Optional[BaseSource] = None,
        datasource_columns: Optional[List[str]] = None,
        datasource_table: Optional[str] = None,
        pod_columns: Optional[List[str]] = None,
        pod_table: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.datasource = datasource
        self._modeller_cols = datasource_columns
        self._modeller_table = datasource_table
        self.pod_columns = pod_columns
        self.pod_table = pod_table
        super().__init__(algorithm=algorithm, **kwargs)

    @classmethod
    def _validate_algorithm(
        cls,
        algorithm: _BaseCompatibleAlgoFactory,
    ) -> None:
        """Checks that `algorithm` is compatible with the protocol."""
        if not isinstance(
            algorithm,
            (_PSICompatibleAlgoFactory_,),
        ):
            raise TypeError(
                f"The {cls.__name__} protocol does not support "
                + f"the {type(algorithm).__name__} algorithm.",
            )

    def dump(self) -> SerializedProtocol:
        """Returns the JSON-serializable representation of the protocol."""
        return cast(SerializedProtocol, bf_dump(self))

    def modeller(self, mailbox: _ModellerMailbox, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the PrivateSetIntersection protocol."""
        algorithm = self.algorithm.modeller()
        return _ModellerSide(
            algorithm=algorithm,
            datasource=self.datasource,
            columns_to_intersect=self._modeller_cols,
            table=self._modeller_table,
            mailbox=mailbox,
            **kwargs,
        )

    def worker(
        self, mailbox: _WorkerMailbox, hub: BitfountHub, **kwargs: Any
    ) -> _WorkerSide:
        """Returns the worker side of the PrivateSetIntersection protocol."""
        return _WorkerSide(
            algorithm=self.algorithm.worker(hub=hub),
            columns_to_intersect=self.pod_columns,
            table=self.pod_table,
            mailbox=mailbox,
            **kwargs,
        )
